# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2016 FUJITSU LIMITED
# Copyright 2017 SUSE Linux GmbH
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Classes for monitoring the monitoring server stack.

    Covering mon-persister, mon-api and mon-thresh.
    Kafka, mysql, vertica and influxdb are covered by other detection plugins. Mon-notification
    uses statsd.
"""

import configparser
import logging
import re
import yaml


import monasca_setup.agent_config
import monasca_setup.detection
from monasca_setup.detection import find_process_cmdline
from monasca_setup.detection.utils import get_agent_username
from monasca_setup.detection.utils import watch_process
from monasca_setup.detection.utils import watch_process_by_username

log = logging.getLogger(__name__)

_APACHE_MARKERS = 'httpd', 'apache',
"""List of all strings in process command line that indicate application
runs in Apache/mod_wsgi"""
_PYTHON_LANG_MARKERS = ('python', 'gunicorn') + _APACHE_MARKERS
"""List of all strings that if found in process exe
mean that application runs under Python"""
_JAVA_LANG_MARKERS = 'java',
"""List of all strings that if found in process exe
mean that application runs under Java"""

_DEFAULT_API_PORT = 8070
"""Default TCP port which monasca-api process should be available by"""


def _get_impl_lang(process):
    """Return implementation language of the application behind process

    :param process: current process
    :type process: psutil.Process
    :return: implementation lang, either java or python
    :rtype: str

    """
    p_exe = process.as_dict(['exe'])['exe']
    for lm in _PYTHON_LANG_MARKERS:
        if lm in p_exe:
            return 'python'
    for lm in _JAVA_LANG_MARKERS:
        if lm in p_exe:
            return 'java'
    raise Exception(('Cannot determine language '
                     'implementation from process exe %s') % p_exe)


class MonAgent(monasca_setup.detection.Plugin):
    """Detect the Monsaca agent engine and setup some simple checks."""

    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        self.available = True
        agent_process_list = ['monasca-collector', 'monasca-forwarder',
                              'monasca-statsd']
        for process in agent_process_list:
            if find_process_cmdline(process) is None:
                self.available = False
                return

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca Agent process check")
        return watch_process_by_username(get_agent_username(), 'monasca-agent',
                                         'monitoring', 'monasca-agent')

    def dependencies_installed(self):
        return True


class MonAPI(monasca_setup.detection.Plugin):
    """Detect mon_api and setup monitoring."""

    PARTIAL_ERR_MSG = 'Plugin for monasca-api will not be configured.'

    def _detect(self):
        """Detects if monasca-api runs in the system

        Method distinguishes between Java and Python implementation
        hence provides different agent's configuration.

        """

        def check_port():
            for conn in api_process.connections('inet'):
                if conn.laddr[1] == api_port:
                    return True
            return False

        def correct_apache_listener(process):
            """Sets api_process to the parent httpd process.

            Method evaluates if process executable is correlated
            with apache-mod_wsgi. If so, retrieves parent process.
            Otherwise returns None

            :param process: current process
            :type process: psutil.Process
            :returns: parent process or None
            :rtype: (psutil.Process, None)

            """

            p_exe = process.as_dict(['exe'])['exe']
            for m in _APACHE_MARKERS:
                if m in p_exe:
                    return process.parent()
            return None

        api_process = find_process_cmdline('monasca-api')
        process_found = api_process is not None

        if process_found:
            impl_lang = _get_impl_lang(api_process)
            if impl_lang == 'python':
                apache_process = correct_apache_listener(api_process)
                if apache_process:
                    log.info('\tmonasca-api runs under Apache WSGI')
                    api_process = apache_process

            impl_helper = self._init_impl_helper(api_process.as_dict(['cmdline'])['cmdline'],
                                                 impl_lang)
            impl_helper.load_configuration()

            api_port = impl_helper.get_bound_port()
            port_taken = check_port()

            if not port_taken:
                log.error('monasca-api is not listening on port %d. %s'
                          % (api_port, self.PARTIAL_ERR_MSG))
                return

            log.info('\tmonasca-api implementation is %s', impl_lang)

            self.available = port_taken
            self._impl_helper = impl_helper
        else:
            log.warning('monasca-api process has not been found. %s'
                        % self.PARTIAL_ERR_MSG)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the monasca-api process check")
        config.merge(watch_process(
            search_strings=['monasca-api'],
            service='monitoring',
            component='monasca-api',
            exact_match=False
        ))
        impl_config = self._impl_helper.build_config()
        if impl_config:
            config.merge(impl_config)

        return config

    def dependencies_installed(self):
        return True

    def _init_impl_helper(self, cmdline, impl_lang):
        """Returns appropriate helper implementation.

        :param impl_lang: implementation language, either `java` or `python`
        :type impl_lang: str
        :return: implementation helper
        :rtype: Union(_MonAPIJavaHelper,_MonAPIPythonHelper)

        """
        if impl_lang == 'java':
            return _MonAPIJavaHelper(cmdline=cmdline)
        else:
            return _MonAPIPythonHelper(cmdline=cmdline, args=self.args)


class MonNotification(monasca_setup.detection.Plugin):
    """Detect the Monsaca notification engine and setup some simple checks."""

    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        if find_process_cmdline('monasca-notification') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca Notification healthcheck")
        notification_process = find_process_cmdline('monasca-notification')
        notification_user = notification_process.as_dict(['username'])['username']
        return watch_process_by_username(notification_user, 'monasca-notification',
                                         'monitoring', 'monasca-notification')

    def dependencies_installed(self):
        return True


class MonPersister(monasca_setup.detection.Plugin):
    """Detect mon_persister and setup monitoring."""

    PARTIAL_ERR_MSG = 'Plugin for monasca-persister will not be configured.'

    def _detect(self):
        """Detects if monasca-persister runs in the system

        Method distinguishes between Java and Python implementation
        hence provides different agent's configuration.

        """
        p_process = find_process_cmdline('monasca-persister')
        process_found = p_process is not None

        if process_found:
            impl_lang = _get_impl_lang(p_process)
            impl_helper = self._init_impl_helper(
                p_process.as_dict(['cmdline'])['cmdline'],
                impl_lang
            )

            if impl_helper is not None:
                impl_helper.load_configuration()

            self._impl_helper = impl_helper
            self.available = True

            log.info('\tmonasca-persister implementation is %s', impl_lang)

        else:
            log.info('monasca-persister process has not been found. %s'
                     % self.PARTIAL_ERR_MSG)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the Monasca persister process check")
        config.merge(watch_process(
            search_strings=['monasca-persister'],
            service='monitoring',
            component='monasca-persister',
            exact_match=False
        ))
        if self._impl_helper is not None:
            impl_config = self._impl_helper.build_config()
            if impl_config:
                config.merge(impl_config)

        return config

    def dependencies_installed(self):
        return True

    @staticmethod
    def _init_impl_helper(cmdline, impl_lang):
        """Returns appropriate helper implementation.

        Note:

            This method returns the helper only for Java.
            Python implementation comes with no extra setup.

        :param impl_lang: implementation language, either `java` or `python`
        :type impl_lang: str
        :return: implementation helper
        :rtype:_MonPersisterJavaHelper

        """
        if impl_lang == 'java':
            return _MonPersisterJavaHelper(cmdline=cmdline)
        return None


class MonThresh(monasca_setup.detection.Plugin):
    """Detect the running mon-thresh and monitor."""

    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        # The node will be running either nimbus or supervisor or both
        self.available = (find_process_cmdline('storm.daemon.nimbus') is not None or
                          find_process_cmdline('storm.daemon.supervisor') is not None)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tWatching the mon-thresh process.")
        config = monasca_setup.agent_config.Plugins()
        for process in ['storm.daemon.nimbus', 'storm.daemon.supervisor', 'storm.daemon.worker']:
            if find_process_cmdline(process) is not None:
                config.merge(
                    watch_process(
                        [process],
                        'monitoring',
                        'apache-storm',
                        exact_match=False,
                        detailed=False))
        config.merge(
            watch_process_by_username(
                'storm',
                'monasca-thresh',
                'monitoring',
                'apache-storm'))
        return config

    def dependencies_installed(self):
        return True


def dropwizard_health_check(service, component, url):
    """Setup a dropwizard heathcheck to be watched by the http_check plugin."""
    config = monasca_setup.agent_config.Plugins()
    config['http_check'] = {'init_config': None,
                            'instances': [
                                {'name': "{0}-{1} healthcheck".format(service, component),
                                 'url': url,
                                 'timeout': 5,
                                 'include_content': False,
                                 'dimensions': {'service': service, 'component': component}}]}
    return config


def dropwizard_metrics(service, component, url, whitelist):
    """Setup a dropwizard metrics check"""
    config = monasca_setup.agent_config.Plugins()
    config['http_metrics'] = {'init_config': None,
                              'instances': [{'name': "{0}-{1} metrics".format(service, component),
                                             'url': url,
                                             'timeout': 5,
                                             'dimensions': {'service': service,
                                                            'component': component},
                                             'whitelist': whitelist}]}
    return config


class _DropwizardJavaHelper(object):
    """Mixing to locate configuration file for DropWizard app

    Class utilizes process of search the configuartion file
    for:

    * monasca-api [**Java**]
    * monasca-persister [**Java**]
    """

    YAML_PATTERN = re.compile(r'.*\.ya?ml', re.IGNORECASE)

    def __init__(self, cmdline=None):
        self._cmdline = cmdline

    def load_configuration(self):
        """Loads java specific configuration.

        Load java specific configuration from:

        * :py:attr:`DEFAULT_CONFIG_FILE`

        :return: True if both configuration files were successfully loaded
        :rtype: bool

        """
        try:
            config_file = self._get_config_file()
            self._read_config_file(config_file)
        except Exception as ex:
            log.error('Failed to parse %s', config_file)
            log.exception(ex)
            raise ex

    def _find_config_file_in_cmdline(self, cmdline):
        # note(trebskit) file should be mentioned
        # somewhere in the end of cmdline
        for item in cmdline[::-1]:
            if self.YAML_PATTERN.match(item):
                return item
        return None

    def _read_config_file(self, config_file):
        with open(config_file, 'r') as config:
            self._cfg = yaml.safe_load(config.read())

    def _get_config_file(self):
        if self._cmdline:
            config_file = self._find_config_file_in_cmdline(
                cmdline=self._cmdline
            )
            if config_file:
                log.debug('\tFound %s for java configuration from CLI',
                          config_file)
                return config_file

        config_file = self.DEFAULT_CONFIG_FILE
        log.debug('\tAssuming default configuration file=%s', config_file)
        return config_file


class _MonPersisterJavaHelper(_DropwizardJavaHelper):
    """Encapsulates Java specific configuration for monasca-persister"""

    DEFAULT_CONFIG_FILE = '/etc/monasca/persister-config.yml'
    """Default location where plugin expects configuration file"""

    def build_config(self):
        config = monasca_setup.agent_config.Plugins()
        metrics = self._collect_metrics()
        self._monitor_endpoints(config, metrics)
        return config

    def _collect_metrics(self):
        """Collects all the metrics .

        Methods will return list of all metrics that will
        later be used in
        :py:mod:`monasca_agent.collector.checks_d.http_metrics` to query
        admin endpoint of monasca-persister.

        Following group of metrics are examined:

        * JVM metrics
        * DB metrics ( see also :py:meth:`._collect_db_metrics` )
        * Internal metrics ( see also :py:meth:`._collect_internal_metrics` )

        :return: list of metrics
        :rtype: list

        """
        log.info("\tEnabling the monasca-persister metrics")

        whitelist = [
            {
                "name": "jvm.memory.total.max",
                "path": "gauges/jvm.memory.total.max/value",
                "type": "gauge"
            },
            {
                "name": "jvm.memory.total.used",
                "path": "gauges/jvm.memory.total.used/value",
                "type": "gauge"
            }
        ]

        self._collect_db_metrics(whitelist)
        self._collect_internal_metrics(whitelist)

        return whitelist

    def _collect_db_metrics(self, whitelist):
        """Collects DB specific metrics.

        Method retrieves which time-series database is used
        in monasca-persister and sets up new metrics to be monitored.

        Note:
            Only if vertica is TSDB in monasca-persister

        """
        database_type = self._cfg['databaseConfiguration']['databaseType']
        if database_type == 'influxdb':
            pass
        elif database_type == 'vertica':
            self._add_vertica_metrics(whitelist)
        else:
            log.warn('Failed finding database type in %s', self.DEFAULT_CONFIG_FILE)

    def _collect_internal_metrics(self, whitelist):
        alarm_num_threads = self._cfg['alarmHistoryConfiguration']['numThreads']
        metric_num_threads = self._cfg['metricConfiguration']['numThreads']

        # Dynamic Whitelist
        for idx in range(alarm_num_threads):
            new_thread = {
                "name": "alarm-state-transitions-added-to-batch-counter[{0}]".format(idx),
                "path": "counters/monasca.persister.pipeline.event."
                        "AlarmStateTransitionHandler[alarm-state-transition-{0}]."
                        "alarm-state-transitions-added-to-batch-counter/count".format(idx),
                "type": "rate"}
            whitelist.append(new_thread)
        for idx in range(metric_num_threads):
            new_thread = {
                "name": "metrics-added-to-batch-counter[{0}]".format(idx),
                "path": "counters/monasca.persister.pipeline.event.MetricHandler[metric-{0}]."
                        "metrics-added-to-batch-counter/count".format(idx),
                "type": "rate"}
            whitelist.append(new_thread)

    def _add_vertica_metrics(self, whitelist):
        whitelist.extend(
            [{"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-cache-hit-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-cache-hit-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-cache-miss-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-cache-miss-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-dimension-cache-hit-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-dimension-cache-hit-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-dimension-cache-miss-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "definition-dimension-cache-miss-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "dimension-cache-hit-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "dimension-cache-hit-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo."
                      "dimension-cache-miss-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "dimension-cache-miss-meter/count",
              "type": "rate"},
             {"name": "monasca.persister.repository.vertica.VerticaMetricRepo.measurement-meter",
              "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo."
                      "measurement-meter/count",
              "type": "rate"}])

    def _monitor_endpoints(self, config, metrics):
        admin_connector = self._cfg['server']['adminConnectors'][0]

        try:
            admin_endpoint_type = admin_connector['type']
        except Exception:
            admin_endpoint_type = "http"
        try:
            admin_endpoint_port = admin_connector['port']
        except Exception:
            admin_endpoint_port = 8091

        log.info("\tEnabling the Monasca persister healthcheck")
        config.merge(
            dropwizard_health_check(
                'monitoring',
                'monasca-persister',
                '{0}://localhost:{1}/healthcheck'.format(admin_endpoint_type,
                                                         admin_endpoint_port)))

        log.info("\tEnabling the Monasca persister metrics")
        config.merge(
            dropwizard_metrics(
                'monitoring',
                'monasca-persister',
                '{0}://localhost:{1}/metrics'.format(admin_endpoint_type,
                                                     admin_endpoint_port),
                metrics))


class _MonAPIJavaHelper(_DropwizardJavaHelper):
    """Encapsulates Java specific configuration for monasca-api"""

    DEFAULT_CONFIG_FILE = '/etc/monasca/api-config.yml'

    def build_config(self):
        """Builds monitoring configuration for monasca-api Java flavour.

        Method configures additional checks that are specific for
        Java implementation of monasca-api.

        List of checks:

        * HttpCheck, :py:mod:`monasca_agent.collector.checks_d.http_check`
        * HttpMetrics, :py:mod:`monasca_agent.collector.checks_d.http_metrics`

        """
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the Monasca api metrics")
        whitelist = [
            {
                "name": "jvm.memory.total.max",
                "path": "gauges/jvm.memory.total.max/value",
                "type": "gauge"
            },
            {
                "name": "jvm.memory.total.used",
                "path": "gauges/jvm.memory.total.used/value",
                "type": "gauge"
            },
            {
                "name": "metrics.published",
                "path": "meters/monasca.api.app.MetricService.metrics.published/count",
                "type": "rate"
            }
        ]

        if not self._is_hibernate_on():
            # if hibernate is not used, it is mysql with DBI
            # for that case having below entries makes sense
            log.debug(
                'monasca-api has not enabled Hibernate, adding DBI metrics')
            whitelist.extend([
                {
                    "name": "raw-sql.time.avg",
                    "path": "timers/org.skife.jdbi.v2.DBI.raw-sql/mean",
                    "type": "gauge"
                },
                {
                    "name": "raw-sql.time.max",
                    "path": "timers/org.skife.jdbi.v2.DBI.raw-sql/max",
                    "type": "gauge"
                }
            ])

        self._monitor_endpoints(config, whitelist)

        return config

    def _monitor_endpoints(self, config, metrics):
        admin_connector = self._cfg['server']['adminConnectors'][0]

        try:
            admin_endpoint_type = admin_connector['type']
        except Exception:
            admin_endpoint_type = "http"
        try:
            admin_endpoint_port = admin_connector['port']
        except Exception:
            admin_endpoint_port = 8081

        healthcheck_url = '{0}://localhost:{1}/healthcheck'.format(
            admin_endpoint_type, admin_endpoint_port)
        metric_url = '{0}://localhost:{1}/metrics'.format(
            admin_endpoint_type, admin_endpoint_port)

        log.info("\tEnabling the monasca-api healthcheck")
        config.merge(dropwizard_health_check('monitoring', 'monasca-api',
                                             healthcheck_url))
        log.info("\tEnabling the monasca-api metrics")
        config.merge(dropwizard_metrics('monitoring', 'monasca-api',
                                        metric_url, metrics))

    def _is_hibernate_on(self):
        # check if api_config has been declared in __init__
        # if not it means that configuration file was not found
        # or monasca-api Python implementation is running

        cfg = getattr(self, '_cfg', None)
        if cfg is None:
            return False

        hibernate_cfg = cfg.get('hibernate', None)
        if hibernate_cfg is None:
            return False

        return hibernate_cfg.get('supportEnabled', False)

    def get_bound_port(self):
        """Returns port API is listening on.

        Method tries to read port from the '/etc/monasca/api-config.yml'
        file. In case if:

        * file was not found in specified location
        * file could be read from the file system
        * file was changed and assumed location of port changed

        code rollbacks to :py:const:`_DEFAULT_API_PORT`

        :return: TCP port api is listening on
        :rtype: int

        """
        if self._cfg is None:
            return _DEFAULT_API_PORT
        try:
            return self._cfg['server']['applicationConnectors'][0]['port']
        except Exception as ex:
            log.error('Failed to read api port from configuration file')
            log.exception(ex)
            return _DEFAULT_API_PORT


class _MonAPIPythonHelper(object):
    """Encapsulates Python specific configuration for monasca-api"""

    DEFAULT_CONFIG_FILE = '/etc/monasca/api-config.ini'
    PASTE_CLI_OPTS = '--paste', '--paster',
    """Possible flags passed to gunicorn processed,
    pointing at paste file"""

    def __init__(self, cmdline=None, args=None):
        super(_MonAPIPythonHelper, self).__init__()
        self._cmdline = cmdline
        self._args = args
        self._paste_config = None

    def build_config(self):
        # note(trebskit) intentionally left empty because gunicorn check
        # seems to have some serious issues with monitoring gunicorn process
        # see https://bugs.launchpad.net/monasca/+bug/1646481
        return None

    def load_configuration(self):
        """Loads INI file from specified path.

        Method loads configuration from specified `path`
        and parses it with :py:class:`configparser.RawConfigParser`

        """
        try:
            config_file = self._get_config_file()
            self._paste_config = self._read_config_file(config_file)
        except Exception as ex:
            log.error('Failed to parse %s', config_file)
            log.exception(ex)
            raise ex

    def get_bound_port(self):
        """Returns port API is listening on.

        Method tries to read port from the '/etc/monasca/api-config.ini'
        file. In case if:

        * file was not found in specified location
        * file could be read from the file system
        * file was changed and assumed location of port changed

        code rollbacks to :py:const:`_DEFAULT_API_PORT`

        :return: TCP port api is listening on
        :rtype: int

        """
        if not self._paste_config:
            return _DEFAULT_API_PORT
        return self._paste_config.getint('server:main', 'port')

    def _read_config_file(self, config_file):
        cp = configparser.RawConfigParser()
        return cp.readfp(open(config_file, 'r'))

    def _get_config_file(self):
        """Method gets configuration file of Python monasca-api.

        Method tries to examine following locations:

        * cmdline of process (looking for either
            of :py:attr:`_MonAPIPythonHelper.PASTE_CLI_OPTS`)
        * this plugin args

        loooking for location of configuration file

        :param args: plugin arguments
        :type args: dict

        """

        if self._cmdline:
            # we're interested in PASTE_CLI_OPTS
            for paste in self.PASTE_CLI_OPTS:
                if paste in self._cmdline:
                    pos = self._cmdline.index(paste)
                    flag = self._cmdline[pos]
                    config_file = self._cmdline[pos + 1]
                    if config_file:
                        log.debug(('\tFound %s=%s for python configuration '
                                   'from CLI'),
                                  flag, config_file)
                        return config_file
                else:
                    log.warn(('\tCannot determine neither %s from process'
                              'cmdline'), self.PASTE_CLI_OPTS)

        if self._args and 'paste-file' in self._args:
            # check if args mentions config file param
            config_file = self._args.get('paste-file', None)
            log.debug(('\tFound paste-file=%s for python configuration '
                       'passed as plugin argument'), config_file)
            return config_file

        config_file = self.DEFAULT_CONFIG_FILE
        log.debug('\tAssuming default paste_file=%s', config_file)

        return config_file



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fcntl
import json
from shutil import rmtree
from socket import gethostname
from unittest import mock
import tempfile
import os
import unittest

from oslo_utils import encodeutils

from monasca_agent.collector.checks_d import json_plugin
import monasca_agent.common.config


def _create_agent_conf():
    # create a temp conf file
    tempdir = tempfile.mkdtemp()
    conf_file = os.path.join(tempdir, 'agent.yaml')
    with open(conf_file, 'wb') as fd:
        fd.write(encodeutils.safe_encode(
            """
            Logging:
              collector_log_file: /var/log/monasca/agent/collector.log
              forwarder_log_file: /var/log/monasca/agent/forwarder.log
              log_level: DEBUG
              statsd_log_file: /var/log/monasca/agent/statsd.log
            Main:
              check_freq: 60
              dimensions: {{}}
              hostname: {hostname}
            """.format(hostname='testhost'), incoming="utf-8")
        )

    config_obj = monasca_agent.common.config.Config(conf_file)
    config = config_obj.get_config(sections='Main')
    # clean up
    rmtree(tempdir, ignore_errors=True)
    return config


fake_now = 1


def FakeNow():
    global fake_now
    return fake_now


class MockJsonPlugin(json_plugin.JsonPlugin):
    def __init__(self):
        super(MockJsonPlugin, self).__init__(
            name='json_plugin',
            init_config={},
            instances=[],
            agent_config=_create_agent_conf()
        )
        self._metrics = []

    @mock.patch("monasca_agent.common.util.get_hostname")
    def check(self, instance, mock_host):
        mock_host.return_value = "testhost"
        self._metrics = []
        return super(MockJsonPlugin, self).check(instance)

    def gauge(self, **kwargs):
        self._metrics.append(kwargs)


def metricsDiffer(expected, actual_orig, ignore_timestamps=True):
    expected = list(expected)
    actual = list(actual_orig)
    if ignore_timestamps:
        for metric in expected:
            metric['timestamp'] = 'ts'
        for metric in actual:
            metric['timestamp'] = 'ts'
    for metric in list(expected):
        if metric not in actual:
            return 'Expected...\n%s\n  ...is missing from actual:\n%s' %\
                   (metrics_sort(metric), metrics_sort(actual_orig))
        actual.remove(metric)
    if actual:
        return 'Unexpected (i.e., extra) metrics:\n%s' % metrics_sort(actual)
    return ''


def metrics_repr(metric):
    m = ''
    for key in ['timestamp', 'metric', 'value', 'dimensions', 'value_meta']:
        m += '%s ' % metric.get(key, '-')
    return m


def metrics_sort(metrics):
    """Makes it easier to debug failed asserts"""
    if isinstance(metrics, list):
        mlist = []
        for metric in metrics:
            mlist.append(metrics_repr(metric))
        mlist.sort()
    else:
        mlist = [metrics_repr(metrics)]
    return '\n'.join(mlist)


def write_metrics_file(file_name, metrics, replace_timestamps=False,
                       stale_age=None):
    file_data = {'replace_timestamps': replace_timestamps,
                 'measurements': []}
    if stale_age:
        file_data.update({'stale_age': stale_age})
    for metric in metrics:
        file_data['measurements'].append(metric)
    with open(file_name, mode='w') as fd:
        fd.write(json.dumps(file_data))


def make_expected(metrics, file_name, now, ts_override=None):
    expected = []
    for metric in list(metrics):
        if ts_override:
            metric['timestamp'] = ts_override
        metric['dimensions'].update({'hostname': 'testhost'})
        expected.append(metric)
    json_plugin_status = {'metric': 'monasca.json_plugin.status', 'value': 0,
                          'dimensions': {'hostname': 'testhost'},
                          'timestamp': now}
    expected.append(json_plugin_status)
    return expected


class JsonPluginCheckTest(unittest.TestCase):
    def setUp(self):
        super(JsonPluginCheckTest, self).setUp()
        self.json_plugin = MockJsonPlugin()

    def test_no_config(self):
        self.json_plugin.check({})

    def test_metric_dir(self):
        tempdir = tempfile.mkdtemp()
        # Empty metrics_dir:
        self.json_plugin.check({'dimensions': {},
                                'metrics_dir': tempdir})
        self.assertEqual([], self.json_plugin.metrics_files)
        expected = [
            {'metric': 'monasca.json_plugin.status', 'value': 0,
             'dimensions': {'hostname': 'testhost'}}]
        differs = metricsDiffer(expected, self.json_plugin._metrics)
        self.assertEqual('', differs, msg=differs)

        # Create json files:
        file1 = os.path.join(tempdir, 'file1.json')
        file2 = os.path.join(tempdir, 'file2.json')
        for metric_file in [file1, file2]:
            with open(metric_file, mode='w') as fd:
                fd.write('[]')
        self.json_plugin.check({'dimensions': {},
                                'metrics_dir': tempdir})
        self.assertIn(file1, self.json_plugin.metrics_files)
        self.assertIn(file2, self.json_plugin.metrics_files)
        rmtree(tempdir, ignore_errors=True)

        expected = [
            {'metric': 'monasca.json_plugin.status', 'value': 0,
             'dimensions': {'hostname': 'testhost'}}
        ]
        differs = metricsDiffer(expected, self.json_plugin._metrics)
        self.assertEqual('', differs, msg=differs)

    def test_bad_json_reporting(self):
        global fake_now
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        with open(file1, mode='w') as fd:
            fd.write('{')
        try:
            with open(file1, 'r') as f:
                json.load(f)
        except (ValueError, TypeError) as e:
            errmsg = 'failed parsing json: %s' %  e
        self.json_plugin.check({'dimensions': {},
                                'metrics_file': file1})
        rmtree(tempdir, ignore_errors=True)
        for now in [1000, 2000]:
            fake_now = now
            expected = [{'metric': 'monasca.json_plugin.status', 'value': 1,
                         'dimensions': {'hostname': 'testhost'},
                         'value_meta': {'msg': '%s: %s' % (file1, errmsg)}}]
            differs = metricsDiffer(expected, self.json_plugin._metrics)
            self.assertEqual('', differs, msg=differs)

    def test_replaced_timestamps(self):
        global fake_now
        json_plugin._now = FakeNow
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        metrics = [
            {'metric': 'name1', 'value': 1,
             'dimensions': {'dim1': 'dim1val'}},
            {'metric': 'name2', 'value': 2,
             'dimensions': {'dim2': 'dim2val'}}
        ]

        write_metrics_file(file1, metrics, replace_timestamps=True)
        for now in [1000, 2000]:
            fake_now = now
            expected = make_expected(metrics, file1, now, ts_override=now)
            self.json_plugin.check({'dimensions': {},
                                   'metrics_file': file1})
            differs = metricsDiffer(expected, self.json_plugin._metrics,
                                    ignore_timestamps=False)
            self.assertEqual('', differs, msg=differs)
        rmtree(tempdir, ignore_errors=True)

    def test_with_timestamps(self):
        global fake_now
        json_plugin._now = FakeNow
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        metrics = [
            {'metric': 'name1', 'value': 1,
             'dimensions': {'dim1': 'dim1val'}},
            {'metric': 'name2', 'value': 2,
             'dimensions': {'dim2': 'dim2val'}}
        ]
        for now in [1000, 2000]:
            fake_now = now
            for metric in metrics:
                metric['timestamp'] = now
            write_metrics_file(file1, metrics, replace_timestamps=False,
                               stale_age=3000)
            expected = make_expected(metrics, file1, now)
            self.json_plugin.check({'dimensions': {},
                                   'metrics_file': file1})
            differs = metricsDiffer(expected, self.json_plugin._metrics,
                                    ignore_timestamps=False)
            self.assertEqual('', differs, msg=differs)
        rmtree(tempdir, ignore_errors=True)

    def test_with_stale_age(self):
        global fake_now
        json_plugin._now = FakeNow
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        metrics = [
            {'metric': 'name1', 'value': 1,
             'dimensions': {'dim1': 'dim1val'}},
            {'metric': 'name2', 'value': 2,
             'dimensions': {'dim2': 'dim2val'}}
        ]
        now = 1000
        fake_now = now
        for metric in metrics:
            metric['timestamp'] = now
        write_metrics_file(file1, metrics, replace_timestamps=False,
                           stale_age=500)
        expected = make_expected(metrics, file1, now, ts_override=now)
        self.json_plugin.check({'dimensions': {},
                               'metrics_file': file1})
        differs = metricsDiffer(expected, self.json_plugin._metrics,
                                ignore_timestamps=False)
        self.assertEqual('', differs, msg=differs)

        # Time moves on, but don't re-write the metrics file
        now = 2000
        fake_now = now
        expected = [{'metric': 'monasca.json_plugin.status', 'value': 1,
                     'dimensions': {'hostname': 'testhost'},
                     'value_meta': {
                         'msg': '%s: Metrics are older than 500 seconds;'
                                ' file not updating?' % file1}}]
        self.json_plugin.check({'dimensions': {},
                               'metrics_file': file1})
        differs = metricsDiffer(expected, self.json_plugin._metrics,
                                ignore_timestamps=True)
        self.assertEqual('', differs, msg=differs)
        rmtree(tempdir, ignore_errors=True)

    def test_no_duplicates(self):
        global fake_now
        json_plugin._now = FakeNow
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        metrics = [
            {'metric': 'name1', 'value': 1,
             'dimensions': {'dim1': 'dim1val'}},
            {'metric': 'name2', 'value': 2,
             'dimensions': {'dim2': 'dim2val'}}
        ]
        now = 1000
        fake_now = now
        for metric in metrics:
            metric['timestamp'] = now
        write_metrics_file(file1, metrics, replace_timestamps=False,
                           stale_age=5000)
        expected = make_expected(metrics, file1, now, ts_override=now)
        self.json_plugin.check({'dimensions': {},
                               'metrics_file': file1})
        differs = metricsDiffer(expected, self.json_plugin._metrics,
                                ignore_timestamps=False)
        self.assertEqual('', differs, msg=differs)

        # Time moves on, but don't re-write the metrics file
        now = 2000
        fake_now = now
        # We don't get the metrics from the file again -- just the plugin
        # status metric
        expected = [{'metric': 'monasca.json_plugin.status', 'value': 0,
                     'dimensions': {'hostname': 'testhost'},
                     'timestamp': now}]
        self.json_plugin.check({'dimensions': {},
                               'metrics_file': file1})
        differs = metricsDiffer(expected, self.json_plugin._metrics,
                                ignore_timestamps=False)
        self.assertEqual('', differs, msg=differs)
        rmtree(tempdir, ignore_errors=True)

    def test_validate_metrics(self):
        metrics = [
            {'metric': 'ok1', 'value': 1},
            {'name': 'ok2', 'value': 2},
            {'metric': 'ok3', 'value': 3, 'dimensions': {}, 'value_meta': {},
             'timestamp': 123},
            {'metric': 'bad1'},
            {'metric': 'bad2', 'junk_key': 'extra'},
            {'value': 1, 'value_meta': {'msg': 'no name or metric key'}},
            {'metric': 'ok4', 'value': 1},
        ]
        valid = self.json_plugin._filter_metrics(metrics, 'dummy.json')
        self.assertTrue('dummy.json' in self.json_plugin.plugin_failures)
        self.assertEqual(4, len(valid))

    def test_posted_metrics_are_purged(self):
        global fake_now
        json_plugin._now = FakeNow
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        metrics = [
            {'metric': 'name1', 'value': 1,
             'dimensions': {'dim1': 'dim1val'}},
            {'metric': 'name2', 'value': 2,
             'dimensions': {'dim2': 'dim2val'}}
        ]
        for now in [1000, 2000, 3000, 4000, 5000, 6000]:
            fake_now = now
            for metric in metrics:
                metric['timestamp'] = now
            write_metrics_file(file1, metrics, replace_timestamps=False,
                               stale_age=2000)
            self.json_plugin.check({'dimensions': {},
                                   'metrics_file': file1})
        for metric in self.json_plugin.posted_metrics[file1]:
            self.assertTrue(metric.get('timestamp', 0) >= 2001, 'not purged')
        self.assertTrue(len(self.json_plugin.posted_metrics[file1]) > 0,
                        'posted metrics not being cached')
        rmtree(tempdir, ignore_errors=True)

    def test_take_lock(self):
        tempdir = tempfile.mkdtemp()
        file1 = os.path.join(tempdir, 'file1.json')
        with open(file1, 'w') as fd_writer:
            with open(file1, 'r') as fd_reader:
                fcntl.flock(fd_writer, fcntl.LOCK_EX | fcntl.LOCK_NB)
                with self.assertRaises(IOError):
                    json_plugin.JsonPlugin._take_shared_lock(fd_reader)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from oslo_utils import importutils

from monasca_setup import agent_config
from monasca_setup import detection
from monasca_setup.detection import utils

LOG = logging.getLogger(__name__)


class InfluxDB(detection.Plugin):
    """Detect InfluxDB and sets up its monitoring"""

    PROC_NAME = 'influxd'
    """Name of the InfluxDB process expected to be found in the system"""
    DEFAULTS = {
        'bind_address': '127.0.0.1',
        'bind_port': 8086
    }
    INFLUXDB_NODE_ARG_NAME = 'influxdb_node'

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """

        proc = utils.find_process_name(self.PROC_NAME)
        process_found = proc is not None

        config_file = self._get_config_file(proc) if process_found else None
        config_file_found = config_file is not None

        dependencies_installed = self.dependencies_installed()

        self.available = (process_found and config_file_found and
                          dependencies_installed)

        if not self.available:
            err_chunks = []
            if not process_found:
                err_chunks.append('\tinfluxdb plugin cannot locate '
                                  '"%s" process.' % self.PROC_NAME)
            elif not config_file_found:
                err_chunks.append('\tinfluxdb plugin cannot locate '
                                  'configuration file.')
            elif not dependencies_installed:
                err_chunks.append('\tinfluxdb plugin requires "toml" '
                                  'to be installed')
            LOG.error('Plugin for InfluxDB will not be configured.\n'
                      'Following issue have to be resolved: %s' %
                      '\n'.join(err_chunks))
        else:
            self._config = self._load_config(config_file)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        LOG.info("\tEnabling the Monasca InfluxDB check")
        config = agent_config.Plugins()

        config.merge(self._monitor_process())
        config.merge(self._monitor_ping_endpoint())

        # NOTE(trebskit) features for next iteration
        # TODO(trebskit) monitor influx metrics (show stats, show diagnostic)
        # TODO(trebskit) monitor certificate expiration

        return config

    def _monitor_process(self):
        dimensions = {}
        if self.args and self.args.get(self.INFLUXDB_NODE_ARG_NAME):
            dimensions.update(
                {self.INFLUXDB_NODE_ARG_NAME: self.args.get(self.INFLUXDB_NODE_ARG_NAME)})

        return detection.watch_process([self.PROC_NAME],
                                       service='influxdb',
                                       component='influxdb',
                                       exact_match=False,
                                       dimensions=dimensions)

    def _monitor_ping_endpoint(self):
        config = agent_config.Plugins()
        http_conf = self._config.get('http', None)

        if http_conf['enabled']:
            host, port = self._explode_bind_address(http_conf)
            listening = utils.find_addrs_listening_on_port(port)

            dimensions = {'service': 'influxdb', 'component': 'influxdb'}
            if self.args and self.args.get(self.INFLUXDB_NODE_ARG_NAME):
                dimensions.update(
                    {self.INFLUXDB_NODE_ARG_NAME: self.args.get(self.INFLUXDB_NODE_ARG_NAME)})

            if listening:
                config['http_check'] = {
                    'init_config': None,
                    'instances': [
                        {
                            'name': 'influxdb',
                            'url': 'http://%s:%d/ping' % (host, port),
                            'dimensions': dimensions
                        }
                    ]
                }
            else:
                LOG.warning('\tInfluxDB[http] is enabled but nothing '
                            'could be found listening at %d port', port)
        else:
            LOG.info('\tInfluxDB[http] is not enabled, skipping')

        return config

    def dependencies_installed(self):
        return importutils.try_import('toml', False)

    @staticmethod
    def _explode_bind_address(http_conf):
        bind_address = http_conf['bind-address']
        path, port = bind_address.split(':')

        if not path:
            path = InfluxDB.DEFAULTS['bind_address']
        if not port:
            port = InfluxDB.DEFAULTS['bind_port']

        return path, int(port)

    @staticmethod
    def _load_config(config_file):
        """Loads toml configuration from specified path.

        Method loads configuration from specified `path`
        and parses it with :py:class:`configparser.RawConfigParser`

        """
        try:
            return importutils.import_module('toml').load(config_file)
        except Exception as ex:
            LOG.error('Failed to parse %s', config_file)
            LOG.exception(ex)

    @staticmethod
    def _get_config_file(proc):
        cmdline = proc.as_dict(attrs=['cmdline'])['cmdline']
        config_flag = '-config'
        if config_flag in cmdline:
            pos = cmdline.index(config_flag)
            return cmdline[pos + 1]
        LOG.warning('%s switch was not found in influxdb cmdline',
                    config_flag)
        return None



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# (C) Copyright 2018 SUSE LLC

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from monasca_setup.detection.plugins.http_check import HttpCheck


class TestHttpCheck(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)

    def test_processs_keystone_global_args(self):
        args = ('keystone_url=https://my.keystone.host.com/identity '
                'keystone_user=foo keystone_password=testpasswd '
                'keystone_project=test_project keystone_project_domain=bar '
                'keystone_user_domain=bar use_keystone=True '
                'url=https://myurl.com')
        plugin = HttpCheck(None, args=args)
        conf = plugin.build_config()
        self.assertTrue('keystone_config' in conf['http_check']['init_config'])
        expected_keystone_config = {
            'keystone_url': 'https://my.keystone.host.com/identity',
            'username': 'foo',
            'password': 'testpasswd',
            'project_name': 'test_project',
            'project_domain_name': 'bar',
            'user_domain_name': 'bar'
        }
        self.assertDictEqual(
            expected_keystone_config,
            conf['http_check']['init_config']['keystone_config'])



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Metric data types
"""
import logging

log = logging.getLogger(__name__)


class Metric(object):
    """A base metric class """

    def __init__(self, name, dimensions, tenant):
        self.metric = {'name': name,
                       'dimensions': dimensions.copy()}

        self.value_meta = None
        self.value = None
        self.timestamp = None
        self.tenant = tenant

    def measurement(self, value, timestamp):
        measurement = self.metric.copy()

        if self.value_meta:
            measurement['value_meta'] = self.value_meta.copy()
        else:
            measurement['value_meta'] = None

        measurement['value'] = value
        measurement['timestamp'] = timestamp * 1000

        envelope = {'measurement': measurement,
                    'tenant_id': self.tenant}

        return envelope

    def sample(self, value, sample_rate, timestamp):
        """Save a sample. """
        raise NotImplementedError()

    def flush(self):
        if self.timestamp is None:
            return []
        if self.value is None:
            log.error('Value of None is not supported, will not send. '
                      'Metric %s with dimensions %s at time %d',
                      self.metric['name'],
                      self.metric['dimensions'],
                      self.timestamp)
            return []

        envelope = self.measurement(self.value, self.timestamp)
        self.timestamp = None
        self.value = None
        return [envelope]


class Gauge(Metric):
    """A metric that tracks a value at particular points in time. """

    def __init__(self, name, dimensions, tenant=None):
        super(Gauge, self).__init__(name, dimensions, tenant)

    def sample(self, value, sample_rate, timestamp):
        self.value = value
        self.timestamp = timestamp


class Counter(Metric):
    """A metric that tracks a counter value. """

    def __init__(self, name, dimensions, tenant=None):
        super(Counter, self).__init__(name, dimensions, tenant)

    def sample(self, value, sample_rate, timestamp):
        try:
            inc = float(value) / sample_rate
            if self.timestamp is None:
                self.value = inc
            else:
                self.value += inc
            self.timestamp = timestamp
        except (TypeError, ValueError):
            log.exception("illegal metric {} value {} sample_rate {}".
                          format(self.metric['name'], value, sample_rate))

    # redefine flush method to make counter an integer when sample rates <> 1.0 used
    def flush(self):
        if self.timestamp:
            self.value = int(self.value)
            return super(Counter, self).flush()
        else:
            return []


class Rate(Metric):
    """Track the rate of metrics over each flush interval """

    def __init__(self, name, dimensions, tenant=None):
        super(Rate, self).__init__(name, dimensions, tenant)
        self.start_value = None
        self.start_timestamp = None

    def sample(self, value, sample_rate, timestamp):
        # set first value if missing
        if self.start_timestamp is None:
            self.start_timestamp = timestamp
            self.start_value = value
        # set second value otherwise
        else:
            self.timestamp = timestamp
            self.value = value

    # redefine flush method to calculate rate from metrics
    def flush(self):
        # need at least two timestamps to determine rate
        # is the second one is missing then the first is kept as start value for
        # the subsequent interval
        if self.start_timestamp is None or self.timestamp is None:
            return []

        delta_t = self.timestamp - self.start_timestamp
        delta_v = self.value - self.start_value
        try:
            rate = delta_v / float(delta_t)
        except ZeroDivisionError:
            log.warning(
                'Conflicting values reported for metric %s with dimensions %s at time %d: (%f, %f)',
                self.metric['name'],
                self.metric['dimensions'],
                self.timestamp,
                self.start_value,
                self.value)

            # skip this measurement, but keep value for next cycle
            self.start_value = self.value
            return []

        # make it start value for next interval (even if it is None!)
        self.start_value = self.value
        self.start_timestamp = self.timestamp

        envelope = self.measurement(rate, self.timestamp)
        self.timestamp = None
        self.value = None
        return [envelope]



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import logging
import os

import monasca_setup.agent_config
import monasca_setup.detection
from monasca_setup.detection.utils import find_process_name

log = logging.getLogger(__name__)

mysql_conf = '/root/.my.cnf'
HOST = 'localhost'
PORT = 3306
SOCKET = '/var/run/mysqld/mysqld.sock'


class MySQL(monasca_setup.detection.Plugin):
    """Detect MySQL daemons and setup configuration to monitor them.

        This plugin needs user/password info for mysql setup.
        It needs either the host ip or socket if using
        the default localhost hostname.  You cannot use
        the localhost name if using ssl.  This plugin
        accepts arguments, and if none are input it will
        try to read the default config file which is
        best placed in /root/.my.cnf in a format such as
        [client]
            user=root
            password=yourpassword
            host=padawan-ccp-c1-m1-mgmt
            ssl_ca=/etc/ssl/certs/ca-certificates.crt

    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        process_exist = find_process_name('mysqld') is not None
        has_dependencies = self.dependencies_installed()
        has_args_or_config_file = (self.args is not None or
                                   os.path.isfile(mysql_conf))
        self.available = (process_exist and has_args_or_config_file and
                          has_dependencies)
        if not self.available:
            if not process_exist:
                log.info('MySQL process does not exist.')
            elif not has_args_or_config_file:
                log.warning(('MySQL process exists but '
                             'configuration file was not found and '
                             'no arguments were given.'))
            elif not has_dependencies:
                log.error(('MySQL process exists but required dependence '
                           'PyMySQL is not installed.'))

    def _get_config(self):
        """Set the configuration to be used for connecting to mysql
        :return:
        """
        self.ssl_options = {}
        # reads default config file if no input parameters
        if self.args is None:
            self._read_config(mysql_conf)
        else:
            self.host = self.args.get('host', HOST)
            self.port = self.args.get('port', PORT)
            self.user = self.args.get('user', 'root')
            self.password = self.args.get('password', None)
            self.socket = self.args.get('socket', None)
            self.ssl_ca = self.args.get('ssl_ca', None)
            self.ssl_key = self.args.get('ssl_key', None)
            self.ssl_cert = self.args.get('ssl_cert', None)
        if self.ssl_ca is not None:
            self.ssl_options['ca'] = self.ssl_ca
        if self.ssl_key is not None:
            self.ssl_options['key'] = self.ssl_key
        if self.ssl_cert is not None:
            self.ssl_options['cert'] = self.ssl_cert
        if self.socket is None and (self.host == 'localhost' or self.host == '127.0.0.1'):
            self.socket = SOCKET

    def _read_config(self, config_file):
        """Read the configuration setting member variables as appropriate.
        :param config_file: The filename of the configuration to read and parse
        """
        log.info("\tUsing client credentials from {}".format(config_file))
        parser = configparser.RawConfigParser(defaults={
            'user': None,
            'password': None,
            'host': HOST,
            'port': PORT,
            'socket': None,
            'ssl_ca': None,
            'ssl_key': None,
            'ssl_cert': None
        }, allow_no_value=True)
        parser.read(config_file)
        self.user = parser.get('client', 'user')
        self.password = parser.get('client', 'password')
        self.host = parser.get('client', 'host')
        self.port = parser.get('client', 'port')
        self.socket = parser.get('client', 'socket')
        self.ssl_ca = parser.get('client', 'ssl_ca')
        self.ssl_key = parser.get('client', 'ssl_key')
        self.ssl_cert = parser.get('client', 'ssl_cert')

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = monasca_setup.agent_config.Plugins()
        # First watch the process
        config.merge(monasca_setup.detection.watch_process(['mysqld'], 'mysql'))
        log.info("\tWatching the mysqld process.")

        try:
            import pymysql
            self._get_config()
            # connection test
            pymysql.connect(host=self.host, user=self.user,
                            passwd=self.password, port=self.port,
                            unix_socket=self.socket, ssl=self.ssl_options)

            log.info("\tConnection test success.")
            config['mysql'] = {
                'init_config': None, 'instances':
                    [{'name': self.host, 'server': self.host, 'port': self.port,
                      'user': self.user, 'pass': self.password,
                      'sock': self.socket, 'ssl_ca': self.ssl_ca,
                      'ssl_key': self.ssl_key, 'ssl_cert': self.ssl_cert}]}
        except ImportError as e:
            exception_msg = ('The mysql dependency PyMySQL is not '
                             'installed. {}'.format(e))
            log.exception(exception_msg)
            raise Exception(exception_msg)
        except pymysql.MySQLError as e:
            exception_msg = 'Could not connect to mysql. {}'.format(e)
            log.exception(exception_msg)
            raise Exception(exception_msg)
        except Exception as e:
            exception_msg = 'Error configuring the mysql check plugin. {}'.format(e)
            log.exception(exception_msg)
            raise Exception(exception_msg)

        return config

    def dependencies_installed(self):
        try:
            import pymysql  # noqa
        except ImportError:
            return False
        return True



from unittest import mock
import unittest

import monasca_agent.common.metrics as metrics_pkg
import monasca_agent.statsd.udp as udp


class TestStatsd(unittest.TestCase):
    def testSubmitPacket(self):
        mock_aggregator = mock.Mock()
        srv = udp.Server(mock_aggregator, 'localhost', 1234)
        test_packet = b"monasca.log.out_logs_truncated_bytes:0|g|#" \
            b"{'service': 'monitoring', 'component': 'monasca-log-api'}"
        srv.submit_packets(test_packet)
        mock_aggregator.submit_metric.assert_called_once_with(
            'monasca.log.out_logs_truncated_bytes',
            0,
            metrics_pkg.Gauge,
            dimensions={
                'service': 'monitoring',
                'component': 'monasca-log-api'},
            sample_rate=1)

    def testSubmitPackets(self):
        mock_aggregator = mock.Mock()
        srv = udp.Server(mock_aggregator, 'localhost', 1234)
        test_packet = b"monasca.log.out_logs_truncated_bytes:0|g|#" \
            b"{'service': 'monitoring', 'component': 'monasca-log-api'}\n" \
            b"application_metric:10|c|#{'service': 'workload'}"
        srv.submit_packets(test_packet)
        mock_aggregator.submit_metric.assert_has_calls([
            mock.call(
                'monasca.log.out_logs_truncated_bytes',
                0,
                metrics_pkg.Gauge,
                dimensions={
                    'service': 'monitoring',
                    'component': 'monasca-log-api'},
                sample_rate=1
            ),
            mock.call(
                'application_metric',
                10,
                metrics_pkg.Counter,
                dimensions={
                    'service': 'workload'},
                sample_rate=1
            )
        ])



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
VCenter plugin that returns only vm status. Takes no instances,
reads from a single configured VCenter
"""

from monasca_agent.collector.checks import AgentCheck
from oslo_vmware import api
from oslo_vmware import vim_util

STATUS_MAP = {
    "connected": 0
}


class VcenterSlim(AgentCheck):
    def __init__(self, name, init_config, agent_config):
        AgentCheck.__init__(self, name, init_config, agent_config, instances=[{}])
        self.max_objects = init_config.get("vcenter_max_objects", 100000)

    def check(self, instance):
        dim_base = self._set_dimensions(None, instance)
        allowed_keys = set(self.init_config.get("allowed_keys", []))
        key_map = self.init_config.get("key_map", {})

        session = self.get_api_session()

        result = session.invoke_api(
            vim_util,
            "get_objects",
            session.vim,
            "VirtualMachine",
            self.max_objects,
            ["runtime.connectionState",
             "config.annotation",
             "config.instanceUuid"])
        for vm in result[0]:
            vm_status = 1
            # vm_name = vm.obj.value
            vm_dimensions = dim_base.copy()
            for prop in vm.propSet:
                if prop.name == "runtime.connectionState":
                    if prop.val in STATUS_MAP:
                        vm_status = STATUS_MAP[prop.val]
                    else:
                        vm_status = 1
                if prop.name == "config.annotation":
                    for line in prop.val.split("\n"):
                        key_val = line.split(":")
                        if len(key_val) == 2 and key_val[0] in allowed_keys:
                            value = key_val[1]
                            key = key_val[0]
                            if key in key_map.keys():
                                key = key_map[key_val[0]]
                            vm_dimensions[key] = value
                if prop.name == "config.instanceUuid":
                    vm_dimensions["instance_id"] = prop.val

            self.gauge("vm.status", vm_status, vm_dimensions)

        session.logout()

    def get_api_session(self):
        api_session = api.VMwareAPISession(
            self.init_config.get("vcenter_ip", ""),
            self.init_config.get("vcenter_user", ""),
            self.init_config.get("vcenter_password", ""),
            self.init_config.get("retry_count", 3),  # retry count
            self.init_config.get("poll_interval", 0.5),  # task_poll_interval
            port=self.init_config.get("vcenter_port", 443),
            scheme="https")
        return api_session



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import unittest


import monasca_agent.collector.checks_d.ib_network as ib_network


class MockIBNetworkPlugin(ib_network.IBNetwork):
    def __init__(self):
        # Don't call the base class constructor
        pass

    @staticmethod
    def _set_dimensions(dimensions, instance=None):
        return {'hostname': 'dummy_hostname'}

    @staticmethod
    def _get_devices():
        return ['mlx5_0', 'mlx5_1']

    @staticmethod
    def _get_fields(device):
        return ['port_rcv_data', 'port_rcv_pkts']


class TestIBNetwork(unittest.TestCase):
    def setUp(self):
        self.ib_network = MockIBNetworkPlugin()

    @mock.patch('monasca_agent.collector.checks_d.ib_network.open',
                mock.mock_open(read_data='1024'))
    @mock.patch('monasca_agent.collector.checks.AgentCheck.rate',
                autospec=True)
    def test_check(self, mock_rate):
        self.ib_network.check(None)
        # For each of the two dummy devices we expect to collect two dummy
        # fields. The count for port_rcv_data should be multiplied by the lane
        # count.
        calls = [
            mock.call(
                mock.ANY,
                ib_network._METRIC_NAME_PREFIX + '.port_rcv_data',
                4096,
                device_name='mlx5_0',
                dimensions={'hostname': 'dummy_hostname'}
            ),
            mock.call(
                mock.ANY,
                ib_network._METRIC_NAME_PREFIX + '.port_rcv_pkts',
                1024,
                device_name='mlx5_0',
                dimensions={'hostname': 'dummy_hostname'}
            ),
            mock.call(
                mock.ANY,
                ib_network._METRIC_NAME_PREFIX + '.port_rcv_data',
                4096,
                device_name='mlx5_1',
                dimensions={'hostname': 'dummy_hostname'}
            ),
            mock.call(
                mock.ANY,
                ib_network._METRIC_NAME_PREFIX + '.port_rcv_pkts',
                1024,
                device_name='mlx5_1',
                dimensions={'hostname': 'dummy_hostname'}
            ),
        ]
        mock_rate.assert_has_calls(calls, any_order=True)



#!/bin/env python
# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Copyright 2017 Fujitsu LIMITED
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from copy import deepcopy
import json
import math
import os
import re
import socket
import stat
import subprocess
import time

from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client

from monasca_agent.collector.checks import AgentCheck
import monasca_agent.collector.checks.utils as utils
from monasca_agent.common import keystone
from monasca_agent import version as ma_version

OVS_CMD = """\
%s --columns=name,external_ids,statistics,options \
--format=json --data=json list Interface\
"""
DPDK_PORT_PREFIX = 'vhu'
HEALTH_METRICS = ['tx_errors', 'rx_errors', 'tx_dropped', 'rx_dropped']

"""Monasca Agent interface for ovs router port and dhcp metrics"""


class OvsCheck(AgentCheck):

    """This check gathers open vswitch router and dhcp metrics.
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        AgentCheck.__init__(self, name, init_config,
                            agent_config, instances=[{}])

        cache_dir = self.init_config.get('cache_dir')
        self.ctr_cache_file = os.path.join(cache_dir, 'ovs_metrics.json')
        self.port_cache_file = os.path.join(cache_dir, 'ovs_ports.json')

        self.use_bits = self.init_config.get('network_use_bits')
        self.check_router_ha = self.init_config.get('check_router_ha')
        self.ovs_cmd = OVS_CMD % self.init_config.get('ovs_cmd')
        include_re = self.init_config.get('included_interface_re', None)
        self.use_absolute_metrics = self.init_config.get('use_absolute_metrics')
        self.use_rate_metrics = self.init_config.get('use_rate_metrics')
        self.use_health_metrics = self.init_config.get('use_health_metrics')
        self.publish_router_capacity = self.init_config.get('publish_router_capacity')
        if include_re is None:
            include_re = 'qg.*'
        else:
            include_re = include_re + '|' + 'qg.*'
        self.include_iface_re = re.compile(include_re)
        self.session = keystone.get_session(**self.init_config)

    def check(self, instance):
        time_start = time.time()
        interface_data = self._get_ovs_data()
        sample_time = float("{:9f}".format(time.time()))
        measure = self._get_measure()

        if not interface_data:
            #
            # No OVS data, nothing to do here.
            #
            return

        port_cache = self._load_port_cache()
        ctr_cache = self._load_counter_cache()

        dims_base = self._set_dimensions({'service': 'networking',
                                          'component': 'ovs'},
                                         instance)

        ifx_deltas = {}
        for ifx in interface_data:
            if not re.match(self.include_iface_re, ifx):
                self.log.debug("include_iface_re {0} does not match with "
                               "ovs-vsctl interface {1} ".format(self.include_iface_re.pattern,
                                                                 ifx))
                continue

            if ifx not in ctr_cache:
                ctr_cache[ifx] = {}
            for metric_name, idx in self._get_metrics_map(measure).items():
                interface_stats_key = self._get_interface_stats_key(idx, metric_name, measure, ifx)
                statistics_dict = interface_data[ifx]['statistics']
                value = statistics_dict[interface_stats_key] \
                    if interface_stats_key in statistics_dict else 0
                if metric_name in ctr_cache[ifx]:
                    cache_time = ctr_cache[ifx][metric_name]['timestamp']
                    time_diff = sample_time - float(cache_time)
                    try:
                        cache_value = ctr_cache[ifx][metric_name]['value']
                        val_diff = (value - cache_value) / time_diff
                    except ZeroDivisionError:
                        #
                        # Being extra safe here, in case we divide by zero
                        # just skip this reading with check below.
                        #
                        val_diff = -1

                    if val_diff < 0:
                        #
                        # Bad value, save current reading and skip
                        #
                        self.log.warn("Ignoring negative router sample for: "
                                      "{0} new value: {1} old value: {2}"
                                      .format(ifx, value,
                                              ctr_cache[ifx][metric_name]))
                        ctr_cache[ifx][metric_name] = {
                            'timestamp': sample_time,
                            'value': value}
                        continue

                    if ifx not in ifx_deltas:
                        uuid = interface_data[ifx]['external_ids']['iface-id']
                        ifx_deltas.update({ifx: {'port_uuid': uuid}})

                    if self.use_bits and 'bytes' in interface_stats_key:
                        val_diff = val_diff * 8

                    ifx_deltas[ifx][interface_stats_key] = val_diff

                #
                # Save the current counter for this metric to the cache
                # for the next time we wake up
                #
                ctr_cache[ifx][metric_name] = {
                    'timestamp': sample_time,
                    'value': value}
        # Done collecting current rates and updating the cache file,
        # let's publish.
        #
        tried_one_update = False
        host_router_max_bw = 0
        active_routers = 0
        for ifx, value in ifx_deltas.items():

            port_uuid = value['port_uuid']
            if port_uuid not in port_cache and not tried_one_update:
                #
                # Only attempt to update port cache
                # file for a missing port uuid once per wakeup.
                #
                tried_one_update = True
                log_msg = "port_uuid {0} not in port cache -- updating."
                self.log.info(log_msg.format(port_uuid))
                port_cache = self._update_port_cache()
            if not port_cache:
                self.log.error("port_cache is empty.")
                continue
            port_info = port_cache.get(port_uuid)
            if not port_info:
                log_msg = "port_uuid {0} not known to neutron -- ghost port?"
                self.log.error(log_msg.format(port_uuid))
                continue
            device_uuid = port_info['device_uuid']
            is_router_port = port_info['is_router_port']
            tenant_id = port_info['tenant_id']
            tenant_name = None
            if 'tenant_name' in port_info:
                tenant_name = port_info['tenant_name']
            if is_router_port and not self._is_active_router(device_uuid):
                continue
            if is_router_port:
                router_name = port_info['router_name']
                if not router_name:
                    log_msg = "Missing router name for router_uuid {0} -- skipping."
                    self.log.error(log_msg.format(device_uuid))
                    continue
                ifx_dimensions = {'resource_id': device_uuid,
                                  'port_id': port_uuid,
                                  'router_name': router_name}
                active_routers += 1
            else:
                ifx_dimensions = {'resource_id': device_uuid,
                                  'port_id': port_uuid}

            this_dimensions = dims_base.copy()
            this_dimensions.update(ifx_dimensions)
            customer_dimensions = this_dimensions.copy()
            del customer_dimensions['hostname']
            ops_dimensions = this_dimensions.copy()
            ops_dimensions.update({'tenant_id': tenant_id})
            if tenant_name:
                ops_dimensions.update({'tenant_name': tenant_name})

            for metric_name, idx in self._get_metrics_map(measure).items():
                # POST to customer project
                interface_stats_key = self._get_interface_stats_key(idx, metric_name, measure, ifx)
                if interface_stats_key not in value:
                    #
                    # If we've skipped a given metric above due to
                    # counter rollover/negative value, we won't have
                    # a value to publish for that metric this round.
                    #
                    continue
                if is_router_port:
                    metric_name_rate = "vrouter.{0}_sec".format(metric_name)
                    metric_name_abs = "vrouter.{0}".format(metric_name)
                else:
                    metric_name_rate = "vswitch.{0}_sec".format(metric_name)
                    metric_name_abs = "vswitch.{0}".format(metric_name)
                if not self.use_health_metrics and interface_stats_key in HEALTH_METRICS:
                    continue
                if self.use_rate_metrics:
                    self.gauge(metric_name_rate, value[interface_stats_key],
                               dimensions=customer_dimensions,
                               delegated_tenant=tenant_id,
                               hostname='SUPPRESS')
                    # POST to operations project with "ovs." prefix
                    self.gauge("ovs.{0}".format(metric_name_rate), value[interface_stats_key],
                               dimensions=ops_dimensions)
                if self.use_absolute_metrics:
                    statistics_dict = interface_data[ifx]['statistics']
                    abs_value = statistics_dict[interface_stats_key] \
                        if interface_stats_key in statistics_dict else 0
                    if self.use_bits and 'bytes' in interface_stats_key:
                        abs_value = abs_value * 8
                    # POST to customer
                    self.gauge(metric_name_abs, abs_value,
                               dimensions=customer_dimensions,
                               delegated_tenant=tenant_id,
                               hostname='SUPPRESS')
                    # POST to operations project
                    self.gauge("ovs.{0}".format(metric_name_abs), abs_value,
                               dimensions=ops_dimensions)

            self._publish_max_bw_metrics(port_info, customer_dimensions,
                                         ops_dimensions)
            host_router_max_bw += self._get_port_cache_max_bw(port_info)

        if host_router_max_bw > 0:
            self.gauge('ovs.vrouter.host_max_bw_kb', host_router_max_bw,
                       dimensions=dims_base)

        self.gauge('ovs.vrouter.active_routers', active_routers, dimensions=dims_base)
        self._update_counter_cache(ctr_cache,
                                   math.ceil(time.time() - time_start), measure)

    def _get_ovs_data(self):

        data_columns = ['name', 'external_ids', 'statistics', 'options']
        output = self._process_command(self.ovs_cmd)

        parsed_ovs_data = {}
        if not output:
            return parsed_ovs_data

        try:
            raw_ovs_json_data = json.loads(output)
        except ValueError:
            #
            # Make sure we got valid json
            #
            return {}

        # There are multiple lines returned, one for each device
        for line in raw_ovs_json_data['data']:

            ifx_data = {}
            ifx_name = None

            # There is one entry in each line for every data_column
            # print "Row:"
            for column_num in range(0, len(data_columns)):
                # Two choices here, it's either a value or a (sub) map
                # If it's a value it becomes a key in the hierarchy
                if line[column_num][0] == "map":
                    map_data = {}
                    for value in line[column_num][1]:
                        map_data.update({value[0]: value[1]})
                    ifx_data.update({data_columns[column_num]: map_data})
                else:
                    # In this specific case name (interface) is our main key
                    # How this would generalize, I don't know.
                    if data_columns[column_num] == "name":
                        ifx_name = line[column_num]

            # Put the interface data in the main dictionary,
            # ensuring we actually built something with the output.
            if ifx_name and ifx_data:
                parsed_ovs_data.update({ifx_name: ifx_data})

        return parsed_ovs_data

    def _get_os_info(self, uuid, all_data):
        for data in all_data:
            if data['id'] == uuid:
                return data
        return None

    def _get_nova_client(self):
        region_name = self.init_config.get('region_name')
        endpoint_type = self.init_config.get("endpoint_type", "publicURL")
        nc = nova_client.Client(2, session=self.session,
                                endpoint_type=endpoint_type,
                                service_type="compute",
                                region_name=region_name,
                                client_name='monasca-agent[ovs]',
                                client_version=ma_version.version_string)

        return nc

    def _get_neutron_client(self):
        region_name = self.init_config.get('region_name')
        endpoint_type = self.init_config.get("endpoint_type", "publicURL")
        return neutron_client.Client(session=self.session,
                                     region_name=region_name,
                                     endpoint_type=endpoint_type,
                                     client_name='monasca-agent[ovs]',
                                     client_version=ma_version.version_string)

    def _run_command(self, command, input=None):
        self.log.debug("Executing command - {0}".format(command))

        errcode = None
        stdout = None
        stderr = None

        try:
            process = subprocess.Popen(command,
                                       shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       stdin=subprocess.PIPE)
            stdout, stderr = process.communicate(input=input)
            errcode = process.returncode

            self.log.debug('errcode - {0}, stdout - {1}, stderr - {2}'.
                           format(errcode,
                                  stdout,
                                  stderr))

        except Exception:
            self.log.error("Failure while executing command - {0}".
                           format(command))

        return errcode, stdout, stderr

    def _process_command(self, command):
        """Runs the command and returns json output
        """
        errcode, stdout, stderr = self._run_command(command)
        cmd_output = ''
        if stdout:
            for line in stdout:
                cmd_output = cmd_output + line
        return cmd_output

    def _load_counter_cache(self):
        """Load the counter metrics from the previous collection iteration
        """
        ctr_cache = {}
        try:
            with open(self.ctr_cache_file, 'r') as cache_json:
                ctr_cache = json.load(cache_json)
        except (IOError, TypeError, ValueError):
            #
            # Couldn't load a cache file, or it's invalid json
            # and is corrupt.  By returning an empty dict we'll ensure
            # that the cache gets rebuilt.
            #
            self.log.warning("Counter cache missing or corrupt, rebuilding.")
            ctr_cache = {}

        return ctr_cache

    def _update_counter_cache(self, ctr_cache, run_time, measure):
        # Remove migrated or deleted ports from the counter cache
        write_ctr_cache = deepcopy(ctr_cache)

        #
        # Grab the first metric name and see if it's in the cache.
        #
        metric_name = list(self._get_metrics_map(measure).keys())[0]

        for ifx in ctr_cache:
            if metric_name not in ctr_cache[ifx]:
                self.log.info("Expiring old/empty {0} from cache".format(ifx))
                del(write_ctr_cache[ifx])
        try:
            with open(self.ctr_cache_file, 'w') as cache_json:
                json.dump(write_ctr_cache, cache_json)
            if stat.S_IMODE(os.stat(self.ctr_cache_file).st_mode) != 0o600:
                os.chmod(self.ctr_cache_file, 0o600)
        except IOError as e:
            self.log.error("Cannot write to {0}: {1}".
                           format(self.ctr_cache_file, e))

    def _get_metrics_map(self, measure):
        metrics_map = {"out_%s" % measure: "tx_bytes",
                       "in_%s" % measure: "rx_bytes",
                       "in_packets": "rx_packets",
                       "out_packets": "tx_packets",
                       "in_dropped": "rx_dropped",
                       "out_dropped": "tx_dropped",
                       "in_errors": "rx_errors",
                       "out_errors": "tx_errors"}

        return metrics_map

    def _update_port_cache(self):
        """Collect port_uuid, device_uuid, router_name, and tenant_id
        for all routers.
        """
        if not hasattr(self, 'neutron_client'):
            self.neutron_client = self._get_neutron_client()
        port_cache = {}

        try:
            self.log.debug("Retrieving Neutron port data")
            all_ports_data = self.neutron_client.list_ports()
            self.log.debug("Retrieving Neutron router data")
            all_routers_data = self.neutron_client.list_routers()
        except Exception as e:
            self.log.exception("Unable to get neutron data: %s", str(e))
            return port_cache

        all_ports_data = all_ports_data['ports']
        all_routers_data = all_routers_data['routers']

        #
        # Only make the keystone call to get the tenant list
        # if we are configured to publish tenant names.
        #
        if self.init_config.get('metadata') and 'tenant_name' in self.init_config.get('metadata'):
            tenants = utils.get_tenant_list(self.init_config, self.log)
        else:
            tenants = []

        for port_data in all_ports_data:
            port_uuid = port_data['id']
            device_uuid = port_data['device_id']
            router_info = self._get_os_info(device_uuid, all_routers_data)
            if router_info:
                tenant_id = router_info['tenant_id']
                is_router_port = True
                router_name = router_info['name']
            else:
                tenant_id = port_data['tenant_id']
                is_router_port = False
                router_name = ""
            port_cache[port_uuid] = {'device_uuid': device_uuid,
                                     'router_name': router_name,
                                     'is_router_port': is_router_port,
                                     'tenant_id': tenant_id}

            tenant_name = utils.get_tenant_name(tenants, tenant_id)
            if tenant_name:
                port_cache[port_uuid]['tenant_name'] = tenant_name

        port_cache = self._add_max_bw_to_port_cache(port_cache,
                                                    all_ports_data)
        port_cache['last_update'] = int(time.time())

        # Write the updated cache
        try:
            with open(self.port_cache_file, 'w') as cache_json:
                json.dump(port_cache, cache_json)
            if stat.S_IMODE(os.stat(self.port_cache_file).st_mode) != 0o600:
                os.chmod(self.port_cache_file, 0o600)
        except IOError as e:
            self.log.error("Cannot write to {0}: {1}".
                           format(self.port_cache_file, e))
        return port_cache

    def _get_port_cache_max_bw(self, port_info):
        if port_info['is_router_port'] and 'max_bw_kb' in port_info:
            return port_info['max_bw_kb']
        else:
            return 0

    def _publish_max_bw_metrics(self, port_info, cust_dims, ops_dims):
        max_bw_kb = self._get_port_cache_max_bw(port_info)

        if not self.publish_router_capacity or max_bw_kb == 0:
            return

        metric_name = 'vrouter.max_bw_kb'

        self.gauge(metric_name, max_bw_kb, dimensions=cust_dims,
                   delegated_tenant=ops_dims['tenant_id'],
                   hostname='SUPPRESS')

        self.gauge("ovs.{0}".format(metric_name), max_bw_kb,
                   dimensions=ops_dims)

    def _get_max_flavor_bw(self, flavor_keys):
        avg_bw = 0
        peak_bw = 0
        burst_bw = 0

        #
        # we'll sum inbound and outbound for the max possible throughput
        #
        avg_re = re.compile('quota:vif_.*bound_average')
        peak_re = re.compile('quota:vif_.*bound_peak')
        burst_re = re.compile('quota:vif_.*bound_burst')

        for key in flavor_keys:
            if re.match(avg_re, key):
                avg_bw += int(flavor_keys[key])
            elif re.match(peak_re, key):
                peak_bw += int(flavor_keys[key])
            elif re.match(burst_re, key):
                burst_bw += int(flavor_keys[key])

        return max(avg_bw, peak_bw, burst_bw)

    def _add_max_bw_to_port_cache(self, port_cache, all_ports_data):
        if not self.publish_router_capacity:
            return port_cache

        tmp_port_cache = deepcopy(port_cache)
        #
        # No need to do a flavor get multiple times
        # for the same flavor when rebuilding the cache.
        #
        flavor_cache = {}

        try:
            if not hasattr(self, 'nova_client'):
                self.nova_client = self._get_nova_client()

            for uuid in port_cache:
                router_max_bw_kb = 0
                port = port_cache[uuid]

                if not port['is_router_port']:
                    continue

                inst_ids = self._get_instance_ids(all_ports_data, port['device_uuid'])

                for instance_id in inst_ids:
                    instance = self.nova_client.servers.get(instance_id)
                    flavor_id = instance.flavor['id']
                    if flavor_id not in flavor_cache:
                        flavor = self.nova_client.flavors.get(instance.flavor['id'])
                        flavor_cache[flavor_id] = flavor.get_keys()
                    router_max_bw_kb += self._get_max_flavor_bw(flavor_cache[flavor_id])

                if router_max_bw_kb > 0:
                    tmp_port_cache[uuid]['max_bw_kb'] = router_max_bw_kb

        except Exception as e:
            msg = "Unable to get the nova instance and flavor info: {0}"
            self.log.error(msg.format(e))

        return tmp_port_cache

    def _get_instance_ids(self, ports, router_uuid):
        subnet_ids = self._get_port_ids(ports,
                                        'network:router_interface',
                                        [router_uuid],
                                        'device_id',
                                        'network_id')

        instance_ids = self._get_port_ids(ports,
                                          'compute:',
                                          subnet_ids,
                                          'network_id',
                                          'device_id')
        return instance_ids

    def _get_port_ids(self, ports, owner, uuids, match_field, return_field):
        return_uuids = []
        if len(uuids) == 0:
            return return_uuids

        for port in ports:
            if port[match_field] in uuids and owner in port['device_owner']:
                return_uuids.append(port[return_field])

        return return_uuids

    def _load_port_cache(self):
        """Load the cache map of router/dhcp port uuids to router uuid, name,
        and tenant name.
        """
        port_cache = {}
        try:
            with open(self.port_cache_file, 'r') as cache_json:
                port_cache = json.load(cache_json)

                # Is it time to force a refresh of this data?
                if self.init_config.get('neutron_refresh') is not None:
                    time_diff = time.time() - port_cache['last_update']
                    if time_diff > self.init_config.get('neutron_refresh'):
                        self.log.warning("Time to update neutron cache file")
                        self._update_port_cache()
        except (IOError, TypeError, ValueError):
            # The file may not exist yet, or is corrupt. Rebuild it now.
            self.log.warning("Port cache doesn't exists , rebuilding.")
            port_cache = self._update_port_cache()

        return port_cache

    def _is_active_router(self, uuid):

        if not self.check_router_ha:
            #
            # We're not configured to check router ha -- let's not bother
            # making the additional neutron calls.
            #
            return True

        active_host = None
        local_host = socket.gethostname()
        try:
            if not hasattr(self, 'neutron_client'):
                self.neutron_client = self._get_neutron_client()
            result = self.neutron_client.list_l3_agent_hosting_routers(uuid)
        except Exception as e:
            #
            # Failed to get the hosting l3 agent, we'll default to calling
            # the router active.
            #
            self.log.error("Unable to get the hosting agent: {0}".format(e))
            return False
        for agent in result['agents']:
            if 'ha_state' not in agent.keys() or agent['ha_state'] is None:
                #
                # HA isn't enabled for this router,
                # so it's active if we've made it this far.
                #
                return True
            if agent['ha_state'] == 'active':
                active_host = agent['host']

        if not active_host:
            #
            # Somehow we didn't find the host above, assume
            # it is active.
            #
            return True

        if active_host != local_host:
            return False

        return True

    def _get_interface_stats_key_for_dpdk(self, metric_name, measure):
        """"Get the interface statistics keys value based on type of port."""
        # The reason for swapping the metric_map value is in DPDK the tx counters
        # given by ovs-vsctl is actually rx counter and vice-versa.
        metrics_map = {"out_%s" % measure: "rx_bytes",
                       "in_%s" % measure: "tx_bytes",
                       "in_packets": "tx_packets",
                       "out_packets": "rx_packets",
                       "out_dropped": "tx_dropped",
                       "in_dropped": "rx_dropped",
                       "out_errors": "tx_errors",
                       "in_errors": "rx_errors"}
        return metrics_map[metric_name]

    def _get_measure(self):
        if self.use_bits:
            measure = 'bits'
        else:
            measure = 'bytes'
        return measure

    def _get_interface_stats_key(self, idx, metric_name, measure, interface):
        """Get the interface statistics key based on the interface."""
        if re.match(r"^" + DPDK_PORT_PREFIX, interface):
            # ovs-vsctl  does not give rx_dropped and tx_error statistics for DPDK ports
            # This check is for DPDK ports.
            return self._get_interface_stats_key_for_dpdk(metric_name,
                                                          measure)
        else:
            # For non DPDK ports return the same idx value.
            return idx



# (C) Copyright 2019 SUSE LLC

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple
from unittest import mock
import unittest

from oslo_config import cfg

from monasca_setup.detection.plugins.libvirt import Libvirt
from monasca_setup.detection import utils


class TestLibvirt(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)

    @mock.patch('monasca_setup.detection.utils.load_oslo_configuration')
    def test_load_oslo_configuration_params(self, patched_load_oslo_config):
        """Test oslo config options.

        Making sure libvirt detection plugin pass the correct
        params to utils.load_oslo_configuration().
        """
        plugin = Libvirt(mock.MagicMock())
        nova_conf = plugin._find_nova_conf(mock.MagicMock())
        self.assertTrue(patched_load_oslo_config.called)
        args, kwargs = patched_load_oslo_config.call_args_list[0]
        self.assertEqual('nova', kwargs['in_project'])
        expected_options = [
            'username',
            'user_domain_name',
            'password',
            'project_name',
            'project_domain_name',
            'auth_url']
        for opt in expected_options:
            self.assertTrue(
                any(d['opt'] == cfg.StrOpt(opt) for d in kwargs['for_opts']))

    def test_init_config(self):
        plugin = Libvirt(mock.MagicMock())
        keystone_authtoken_args = {
            'username': 'foo',
            'user_domain_name': 'Default',
            'password': 'secrete',
            'project_name': 'bar',
            'project_domain_name': 'Default',
            'auth_url': 'https://127.0.0.1:5000/v3'
        }
        plugin.nova_conf = {'keystone_authtoken': keystone_authtoken_args}
        init_config = plugin._get_init_config()
        for arg in keystone_authtoken_args:
            self.assertEqual(keystone_authtoken_args[arg], init_config[arg])



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase
from eventlet.green import time

from monasca_agent.collector.checks.services_checks import ServicesCheck

from monasca_agent.collector.checks.services_checks import Status

class DummyServiceCheck(ServicesCheck):
    def __init__(self, name, init_config, agent_config, instances=None):
        super(DummyServiceCheck, self).__init__(name, init_config, agent_config, instances)

    def _check(self, instance):
        w = instance.get('service_wait', 5)
        time.sleep(w)
        return Status.UP, "UP"

    def set_timeout(self, timeout):
        self.timeout = timeout

class TestServicesCheck(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        pass

    def test_service_check_timeout(self):

        init_config = {}
        agent_config = {'timeout': 4}
        instances = []
        for wait in [1,2, 6, 8]:
            instances.append({'service_wait': wait, 'name' : 'dummy %d' % wait})
        self.dummy_service = DummyServiceCheck("dummy service", init_config, agent_config, instances=instances)
        self.dummy_service.run()
        time.sleep(10)
        self.assertEqual(self.dummy_service.statuses['dummy 1'][0], Status.UP)
        self.assertEqual(self.dummy_service.statuses['dummy 2'][0], Status.UP)

        self.assertEqual(self.dummy_service.statuses['dummy 6'][0], "FAILURE")
        self.assertEqual(self.dummy_service.statuses['dummy 8'][0], "FAILURE")




# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from monasca_setup.detection.args_plugin import ArgsPlugin  # noqa
from monasca_setup.detection.plugin import Plugin  # noqa
from monasca_setup.detection.service_plugin import ServicePlugin  # noqa
from monasca_setup.detection.utils import find_process_cmdline  # noqa
from monasca_setup.detection.utils import find_process_name  # noqa
from monasca_setup.detection.utils import find_process_service  # noqa
from monasca_setup.detection.utils import watch_process  # noqa
from monasca_setup.detection.utils import watch_process_by_username  # noqa



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import time
import unittest

from tests.common import get_check


class VarnishTestCase(unittest.TestCase):

    def setUp(self):
        self.v_dump = """client_conn              42561         1.84 Client connections accepted
client_drop                  0         0.00 Connection dropped, no sess/wrk
client_req             1564940        67.82 Client requests received
cache_hit                52397         2.27 Cache hits
cache_hitpass                0         0.00 Cache hits for pass
cache_miss                5395         0.23 Cache misses
backend_conn             20191         0.88 Backend conn. success
backend_unhealthy            0         0.00 Backend conn. not attempted
backend_busy                 0         0.00 Backend conn. too many
backend_fail              1717         0.07 Backend conn. failures
backend_reuse          1490649        64.60 Backend conn. reuses
backend_toolate            771         0.03 Backend conn. was closed
backend_recycle        1491423        64.64 Backend conn. recycles
backend_retry               13         0.00 Backend conn. retry
fetch_head                   5         0.00 Fetch head
fetch_length           1052795        45.63 Fetch with Length
fetch_chunked           457998        19.85 Fetch chunked
fetch_eof                    0         0.00 Fetch EOF
fetch_bad                    0         0.00 Fetch had bad headers
fetch_close                 13         0.00 Fetch wanted close
fetch_oldhttp                0         0.00 Fetch pre HTTP/1.1 closed
fetch_zero                   0         0.00 Fetch zero len
fetch_failed                 0         0.00 Fetch failed
fetch_1xx                    0         0.00 Fetch no body (1xx)
fetch_204                    0         0.00 Fetch no body (204)
fetch_304                    0         0.00 Fetch no body (304)
n_sess_mem                 547          .   N struct sess_mem
n_sess                      29          .   N struct sess
n_object                   149          .   N struct object
n_vampireobject              0          .   N unresurrected objects
n_objectcore               168          .   N struct objectcore
n_objecthead               211          .   N struct objecthead
n_waitinglist              980          .   N struct waitinglist
n_vbc                        3          .   N struct vbc
n_wrk                       21          .   N worker threads
n_wrk_create              1999         0.09 N worker threads created
n_wrk_failed                 0         0.00 N worker threads not created
n_wrk_max               313745        13.60 N worker threads limited
n_wrk_lqueue                 0         0.00 work request queue length
n_wrk_queued              7125         0.31 N queued work requests
n_wrk_drop                   0         0.00 N dropped work requests
n_backend                    1          .   N backends
n_expired                 5237          .   N expired objects
n_lru_nuked                  0          .   N LRU nuked objects
n_lru_moved              34482          .   N LRU moved objects
losthdr                      0         0.00 HTTP header overflows
n_objsendfile                0         0.00 Objects sent with sendfile
n_objwrite             1611740        69.85 Objects sent with write
n_objoverflow                0         0.00 Objects overflowing workspace
s_sess                   42561         1.84 Total Sessions
s_req                  1564940        67.82 Total Requests
s_pipe                       0         0.00 Total pipe
s_pass                 1507148        65.32 Total pass
s_fetch                1510811        65.48 Total fetch
s_hdrbytes           462276984     20034.54 Total header bytes
s_bodybytes        26560298859   1151092.09 Total body bytes
sess_closed               2198         0.10 Session Closed
sess_pipeline                0         0.00 Session Pipeline
sess_readahead               0         0.00 Session Read Ahead
sess_linger            1563076        67.74 Session Linger
sess_herd              1285271        55.70 Session herd
shm_records          114040454      4942.38 SHM records
shm_writes             6024350       261.09 SHM writes
shm_flushes                  0         0.00 SHM flushes due to overflow
shm_cont                  1353         0.06 SHM MTX contention
shm_cycles                  46         0.00 SHM cycles through buffer
sms_nreq                  1732         0.08 SMS allocator requests
sms_nobj                     0          .   SMS outstanding allocations
sms_nbytes                   0          .   SMS outstanding bytes
sms_balloc              723976          .   SMS bytes allocated
sms_bfree               723976          .   SMS bytes freed
backend_req            1510831        65.48 Backend requests made
n_vcl                        1         0.00 N vcl total
n_vcl_avail                  1         0.00 N vcl available
n_vcl_discard                0         0.00 N vcl discarded
n_ban                        1          .   N total active bans
n_ban_add                    1         0.00 N new bans added
n_ban_retire                 0         0.00 N old bans deleted
n_ban_obj_test               0         0.00 N objects tested
n_ban_re_test                0         0.00 N regexps tested against
n_ban_dups                   0         0.00 N duplicate bans removed
hcb_nolock               57795         2.50 HCB Lookups without lock
hcb_lock                  3909         0.17 HCB Lookups with lock
hcb_insert                3909         0.17 HCB Inserts
esi_errors                   0         0.00 ESI parse errors (unlock)
esi_warnings                 0         0.00 ESI parse warnings (unlock)
accept_fail                  0         0.00 Accept failures
client_drop_late             0         0.00 Connection dropped late
uptime                   23074         1.00 Client uptime
dir_dns_lookups              0         0.00 DNS director lookups
dir_dns_failed               0         0.00 DNS director failed lookups
dir_dns_hit                  0         0.00 DNS director cached lookups hit
dir_dns_cache_full           0         0.00 DNS director full dnscache
vmods                        0          .   Loaded VMODs
n_gzip                       0         0.00 Gzip operations
n_gunzip                     0         0.00 Gunzip operations
LCK.sms.creat                1         0.00 Created locks
LCK.sms.destroy              0         0.00 Destroyed locks
LCK.sms.locks             5196         0.23 Lock Operations
LCK.sms.colls                0         0.00 Collisions
LCK.smp.creat                0         0.00 Created locks
LCK.smp.destroy              0         0.00 Destroyed locks
LCK.smp.locks                0         0.00 Lock Operations
LCK.smp.colls                0         0.00 Collisions
LCK.sma.creat                1         0.00 Created locks
LCK.sma.destroy              0         0.00 Destroyed locks
LCK.sma.locks          6474775       280.61 Lock Operations
LCK.sma.colls                0         0.00 Collisions
LCK.smf.creat                1         0.00 Created locks
LCK.smf.destroy              0         0.00 Destroyed locks
LCK.smf.locks            30590         1.33 Lock Operations
LCK.smf.colls                0         0.00 Collisions
LCK.hsl.creat                0         0.00 Created locks
LCK.hsl.destroy              0         0.00 Destroyed locks
LCK.hsl.locks                0         0.00 Lock Operations
LCK.hsl.colls                0         0.00 Collisions
LCK.hcb.creat                1         0.00 Created locks
LCK.hcb.destroy              0         0.00 Destroyed locks
LCK.hcb.locks             7801         0.34 Lock Operations
LCK.hcb.colls                0         0.00 Collisions
LCK.hcl.creat                0         0.00 Created locks
LCK.hcl.destroy              0         0.00 Destroyed locks
LCK.hcl.locks                0         0.00 Lock Operations
LCK.hcl.colls                0         0.00 Collisions
LCK.vcl.creat                1         0.00 Created locks
LCK.vcl.destroy              0         0.00 Destroyed locks
LCK.vcl.locks             3987         0.17 Lock Operations
LCK.vcl.colls                0         0.00 Collisions
LCK.stat.creat               1         0.00 Created locks
LCK.stat.destroy             0         0.00 Destroyed locks
LCK.stat.locks             547         0.02 Lock Operations
LCK.stat.colls               0         0.00 Collisions
LCK.sessmem.creat            1         0.00 Created locks
LCK.sessmem.destroy            0         0.00 Destroyed locks
LCK.sessmem.locks          43217         1.87 Lock Operations
LCK.sessmem.colls              0         0.00 Collisions
LCK.wstat.creat                1         0.00 Created locks
LCK.wstat.destroy              0         0.00 Destroyed locks
LCK.wstat.locks            74248         3.22 Lock Operations
LCK.wstat.colls                0         0.00 Collisions
LCK.herder.creat               1         0.00 Created locks
LCK.herder.destroy             0         0.00 Destroyed locks
LCK.herder.locks            5250         0.23 Lock Operations
LCK.herder.colls               0         0.00 Collisions
LCK.wq.creat                   2         0.00 Created locks
LCK.wq.destroy                 0         0.00 Destroyed locks
LCK.wq.locks             2618934       113.50 Lock Operations
LCK.wq.colls                   0         0.00 Collisions
LCK.objhdr.creat            4840         0.21 Created locks
LCK.objhdr.destroy          4629         0.20 Destroyed locks
LCK.objhdr.locks          243135        10.54 Lock Operations
LCK.objhdr.colls               0         0.00 Collisions
LCK.exp.creat                  1         0.00 Created locks
LCK.exp.destroy                0         0.00 Destroyed locks
LCK.exp.locks              33597         1.46 Lock Operations
LCK.exp.colls                  0         0.00 Collisions
LCK.lru.creat                  2         0.00 Created locks
LCK.lru.destroy                0         0.00 Destroyed locks
LCK.lru.locks               5386         0.23 Lock Operations
LCK.lru.colls                  0         0.00 Collisions
LCK.cli.creat                  1         0.00 Created locks
LCK.cli.destroy                0         0.00 Destroyed locks
LCK.cli.locks               7696         0.33 Lock Operations
LCK.cli.colls                  0         0.00 Collisions
LCK.ban.creat                  1         0.00 Created locks
LCK.ban.destroy                0         0.00 Destroyed locks
LCK.ban.locks              33611         1.46 Lock Operations
LCK.ban.colls                  0         0.00 Collisions
LCK.vbp.creat                  1         0.00 Created locks
LCK.vbp.destroy                0         0.00 Destroyed locks
LCK.vbp.locks                  0         0.00 Lock Operations
LCK.vbp.colls                  0         0.00 Collisions
LCK.vbe.creat                  1         0.00 Created locks
LCK.vbe.destroy                0         0.00 Destroyed locks
LCK.vbe.locks              43813         1.90 Lock Operations
LCK.vbe.colls                  0         0.00 Collisions
LCK.backend.creat              1         0.00 Created locks
LCK.backend.destroy            0         0.00 Destroyed locks
LCK.backend.locks        3050125       132.19 Lock Operations
LCK.backend.colls              0         0.00 Collisions
SMF.s0.c_req               12925         0.56 Allocator requests
SMF.s0.c_fail                  0         0.00 Allocator failures
SMF.s0.c_bytes         812273664     35202.98 Bytes allocated
SMF.s0.c_freed         799502336     34649.49 Bytes freed
SMF.s0.g_alloc               311          .   Allocations outstanding
SMF.s0.g_bytes          12771328          .   Bytes outstanding
SMF.s0.g_space         524099584          .   Bytes available
SMF.s0.g_smf                 383          .   N struct smf
SMF.s0.g_smf_frag             60          .   N small free smf
SMF.s0.g_smf_large            12          .   N large free smf
SMA.Transient.c_req      3010912       130.49 Allocator requests
SMA.Transient.c_fail           0         0.00 Allocator failures
SMA.Transient.c_bytes  63026331110   2731487.00 Bytes allocated
SMA.Transient.c_freed  63026331110   2731487.00 Bytes freed
SMA.Transient.g_alloc            0          .   Allocations outstanding
SMA.Transient.g_bytes            0          .   Bytes outstanding
SMA.Transient.g_space            0          .   Bytes available
VBE.default(127.0.0.1,,8080).vcls            1          .   VCL references
VBE.default(127.0.0.1,,8080).happy           0          .   Happy health probes"""

        self.xml_dump = """<?xml version="1.0"?>
        <varnishstat timestamp="2012-05-21T17:19:59">
            <stat>
                <name>client_conn</name>
                <value>475607</value>
                <flag>a</flag>
                <description>Client connections accepted</description>
            </stat>
            <stat>
                <name>client_drop</name>
                <value>0</value>
                <flag>a</flag>
                <description>Connection dropped, no sess/wrk</description>
            </stat>
            <stat>
                <name>client_req</name>
                <value>46007888</value>
                <flag>a</flag>
                <description>Client requests received</description>
            </stat>
            <stat>
                <name>cache_hit</name>
                <value>29652727</value>
                <flag>a</flag>
                <description>Cache hits</description>
            </stat>
            <stat>
                <name>cache_hitpass</name>
                <value>0</value>
                <flag>a</flag>
                <description>Cache hits for pass</description>
            </stat>
            <stat>
                <name>cache_miss</name>
                <value>11551328</value>
                <flag>a</flag>
                <description>Cache misses</description>
            </stat>
            <stat>
                <name>backend_conn</name>
                <value>187918</value>
                <flag>a</flag>
                <description>Backend conn. success</description>
            </stat>
            <stat>
                <name>backend_unhealthy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Backend conn. not attempted</description>
            </stat>
            <stat>
                <name>backend_busy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Backend conn. too many</description>
            </stat>
            <stat>
                <name>backend_fail</name>
                <value>912</value>
                <flag>a</flag>
                <description>Backend conn. failures</description>
            </stat>
            <stat>
                <name>backend_reuse</name>
                <value>16078929</value>
                <flag>a</flag>
                <description>Backend conn. reuses</description>
            </stat>
            <stat>
                <name>backend_toolate</name>
                <value>13560</value>
                <flag>a</flag>
                <description>Backend conn. was closed</description>
            </stat>
            <stat>
                <name>backend_recycle</name>
                <value>16092503</value>
                <flag>a</flag>
                <description>Backend conn. recycles</description>
            </stat>
            <stat>
                <name>backend_retry</name>
                <value>45</value>
                <flag>a</flag>
                <description>Backend conn. retry</description>
            </stat>
            <stat>
                <name>fetch_head</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch head</description>
            </stat>
            <stat>
                <name>fetch_length</name>
                <value>12638932</value>
                <flag>a</flag>
                <description>Fetch with Length</description>
            </stat>
            <stat>
                <name>fetch_chunked</name>
                <value>3627795</value>
                <flag>a</flag>
                <description>Fetch chunked</description>
            </stat>
            <stat>
                <name>fetch_eof</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch EOF</description>
            </stat>
            <stat>
                <name>fetch_bad</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch had bad headers</description>
            </stat>
            <stat>
                <name>fetch_close</name>
                <value>36</value>
                <flag>a</flag>
                <description>Fetch wanted close</description>
            </stat>
            <stat>
                <name>fetch_oldhttp</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch pre HTTP/1.1 closed</description>
            </stat>
            <stat>
                <name>fetch_zero</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch zero len</description>
            </stat>
            <stat>
                <name>fetch_failed</name>
                <value>9</value>
                <flag>a</flag>
                <description>Fetch failed</description>
            </stat>
            <stat>
                <name>fetch_1xx</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch no body (1xx)</description>
            </stat>
            <stat>
                <name>fetch_204</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch no body (204)</description>
            </stat>
            <stat>
                <name>fetch_304</name>
                <value>0</value>
                <flag>a</flag>
                <description>Fetch no body (304)</description>
            </stat>
            <stat>
                <name>n_sess_mem</name>
                <value>334</value>
                <flag>i</flag>
                <description>N struct sess_mem</description>
            </stat>
            <stat>
                <name>n_sess</name>
                <value>31</value>
                <flag>i</flag>
                <description>N struct sess</description>
            </stat>
            <stat>
                <name>n_object</name>
                <value>391072</value>
                <flag>i</flag>
                <description>N struct object</description>
            </stat>
            <stat>
                <name>n_vampireobject</name>
                <value>0</value>
                <flag>i</flag>
                <description>N unresurrected objects</description>
            </stat>
            <stat>
                <name>n_objectcore</name>
                <value>391086</value>
                <flag>i</flag>
                <description>N struct objectcore</description>
            </stat>
            <stat>
                <name>n_objecthead</name>
                <value>119540</value>
                <flag>i</flag>
                <description>N struct objecthead</description>
            </stat>
            <stat>
                <name>n_waitinglist</name>
                <value>42255</value>
                <flag>i</flag>
                <description>N struct waitinglist</description>
            </stat>
            <stat>
                <name>n_vbc</name>
                <value>13</value>
                <flag>i</flag>
                <description>N struct vbc</description>
            </stat>
            <stat>
                <name>n_wrk</name>
                <value>24</value>
                <flag>i</flag>
                <description>N worker threads</description>
            </stat>
            <stat>
                <name>n_wrk_create</name>
                <value>3718</value>
                <flag>a</flag>
                <description>N worker threads created</description>
            </stat>
            <stat>
                <name>n_wrk_failed</name>
                <value>0</value>
                <flag>a</flag>
                <description>N worker threads not created</description>
            </stat>
            <stat>
                <name>n_wrk_max</name>
                <value>0</value>
                <flag>a</flag>
                <description>N worker threads limited</description>
            </stat>
            <stat>
                <name>n_wrk_lqueue</name>
                <value>0</value>
                <flag>a</flag>
                <description>work request queue length</description>
            </stat>
            <stat>
                <name>n_wrk_queued</name>
                <value>37387</value>
                <flag>a</flag>
                <description>N queued work requests</description>
            </stat>
            <stat>
                <name>n_wrk_drop</name>
                <value>0</value>
                <flag>a</flag>
                <description>N dropped work requests</description>
            </stat>
            <stat>
                <name>n_backend</name>
                <value>17</value>
                <flag>i</flag>
                <description>N backends</description>
            </stat>
            <stat>
                <name>n_expired</name>
                <value>5509942</value>
                <flag>i</flag>
                <description>N expired objects</description>
            </stat>
            <stat>
                <name>n_lru_nuked</name>
                <value>5650303</value>
                <flag>i</flag>
                <description>N LRU nuked objects</description>
            </stat>
            <stat>
                <name>n_lru_moved</name>
                <value>14960592</value>
                <flag>i</flag>
                <description>N LRU moved objects</description>
            </stat>
            <stat>
                <name>losthdr</name>
                <value>0</value>
                <flag>a</flag>
                <description>HTTP header overflows</description>
            </stat>
            <stat>
                <name>n_objsendfile</name>
                <value>0</value>
                <flag>a</flag>
                <description>Objects sent with sendfile</description>
            </stat>
            <stat>
                <name>n_objwrite</name>
                <value>46952631</value>
                <flag>a</flag>
                <description>Objects sent with write</description>
            </stat>
            <stat>
                <name>n_objoverflow</name>
                <value>0</value>
                <flag>a</flag>
                <description>Objects overflowing workspace</description>
            </stat>
            <stat>
                <name>s_sess</name>
                <value>475606</value>
                <flag>a</flag>
                <description>Total Sessions</description>
            </stat>
            <stat>
                <name>s_req</name>
                <value>46007888</value>
                <flag>a</flag>
                <description>Total Requests</description>
            </stat>
            <stat>
                <name>s_pipe</name>
                <value>0</value>
                <flag>a</flag>
                <description>Total pipe</description>
            </stat>
            <stat>
                <name>s_pass</name>
                <value>4715470</value>
                <flag>a</flag>
                <description>Total pass</description>
            </stat>
            <stat>
                <name>s_fetch</name>
                <value>16266754</value>
                <flag>a</flag>
                <description>Total fetch</description>
            </stat>
            <stat>
                <name>s_hdrbytes</name>
                <value>13489812656</value>
                <flag>a</flag>
                <description>Total header bytes</description>
            </stat>
            <stat>
                <name>s_bodybytes</name>
                <value>870366768630</value>
                <flag>a</flag>
                <description>Total body bytes</description>
            </stat>
            <stat>
                <name>sess_closed</name>
                <value>90657</value>
                <flag>a</flag>
                <description>Session Closed</description>
            </stat>
            <stat>
                <name>sess_pipeline</name>
                <value>0</value>
                <flag>a</flag>
                <description>Session Pipeline</description>
            </stat>
            <stat>
                <name>sess_readahead</name>
                <value>0</value>
                <flag>a</flag>
                <description>Session Read Ahead</description>
            </stat>
            <stat>
                <name>sess_linger</name>
                <value>45917694</value>
                <flag>a</flag>
                <description>Session Linger</description>
            </stat>
            <stat>
                <name>sess_herd</name>
                <value>15876611</value>
                <flag>a</flag>
                <description>Session herd</description>
            </stat>
            <stat>
                <name>shm_records</name>
                <value>2312772697</value>
                <flag>a</flag>
                <description>SHM records</description>
            </stat>
            <stat>
                <name>shm_writes</name>
                <value>97762309</value>
                <flag>a</flag>
                <description>SHM writes</description>
            </stat>
            <stat>
                <name>shm_flushes</name>
                <value>3</value>
                <flag>a</flag>
                <description>SHM flushes due to overflow</description>
            </stat>
            <stat>
                <name>shm_cont</name>
                <value>31213</value>
                <flag>a</flag>
                <description>SHM MTX contention</description>
            </stat>
            <stat>
                <name>shm_cycles</name>
                <value>967</value>
                <flag>a</flag>
                <description>SHM cycles through buffer</description>
            </stat>
            <stat>
                <name>sms_nreq</name>
                <value>88407</value>
                <flag>a</flag>
                <description>SMS allocator requests</description>
            </stat>
            <stat>
                <name>sms_nobj</name>
                <value>0</value>
                <flag>i</flag>
                <description>SMS outstanding allocations</description>
            </stat>
            <stat>
                <name>sms_nbytes</name>
                <value>0</value>
                <flag>i</flag>
                <description>SMS outstanding bytes</description>
            </stat>
            <stat>
                <name>sms_balloc</name>
                <value>477828</value>
                <flag>i</flag>
                <description>SMS bytes allocated</description>
            </stat>
            <stat>
                <name>sms_bfree</name>
                <value>477828</value>
                <flag>i</flag>
                <description>SMS bytes freed</description>
            </stat>
            <stat>
                <name>backend_req</name>
                <value>16266804</value>
                <flag>a</flag>
                <description>Backend requests made</description>
            </stat>
            <stat>
                <name>n_vcl</name>
                <value>1</value>
                <flag>a</flag>
                <description>N vcl total</description>
            </stat>
            <stat>
                <name>n_vcl_avail</name>
                <value>1</value>
                <flag>a</flag>
                <description>N vcl available</description>
            </stat>
            <stat>
                <name>n_vcl_discard</name>
                <value>0</value>
                <flag>a</flag>
                <description>N vcl discarded</description>
            </stat>
            <stat>
                <name>n_ban</name>
                <value>1</value>
                <flag>i</flag>
                <description>N total active bans</description>
            </stat>
            <stat>
                <name>n_ban_add</name>
                <value>1</value>
                <flag>a</flag>
                <description>N new bans added</description>
            </stat>
            <stat>
                <name>n_ban_retire</name>
                <value>0</value>
                <flag>a</flag>
                <description>N old bans deleted</description>
            </stat>
            <stat>
                <name>n_ban_obj_test</name>
                <value>0</value>
                <flag>a</flag>
                <description>N objects tested</description>
            </stat>
            <stat>
                <name>n_ban_re_test</name>
                <value>0</value>
                <flag>a</flag>
                <description>N regexps tested against</description>
            </stat>
            <stat>
                <name>n_ban_dups</name>
                <value>0</value>
                <flag>a</flag>
                <description>N duplicate bans removed</description>
            </stat>
            <stat>
                <name>hcb_nolock</name>
                <value>41204059</value>
                <flag>a</flag>
                <description>HCB Lookups without lock</description>
            </stat>
            <stat>
                <name>hcb_lock</name>
                <value>3053198</value>
                <flag>a</flag>
                <description>HCB Lookups with lock</description>
            </stat>
            <stat>
                <name>hcb_insert</name>
                <value>3053197</value>
                <flag>a</flag>
                <description>HCB Inserts</description>
            </stat>
            <stat>
                <name>esi_errors</name>
                <value>0</value>
                <flag>a</flag>
                <description>ESI parse errors (unlock)</description>
            </stat>
            <stat>
                <name>esi_warnings</name>
                <value>0</value>
                <flag>a</flag>
                <description>ESI parse warnings (unlock)</description>
            </stat>
            <stat>
                <name>accept_fail</name>
                <value>0</value>
                <flag>a</flag>
                <description>Accept failures</description>
            </stat>
            <stat>
                <name>client_drop_late</name>
                <value>0</value>
                <flag>a</flag>
                <description>Connection dropped late</description>
            </stat>
            <stat>
                <name>uptime</name>
                <value>265069</value>
                <flag>a</flag>
                <description>Client uptime</description>
            </stat>
            <stat>
                <name>dir_dns_lookups</name>
                <value>0</value>
                <flag>a</flag>
                <description>DNS director lookups</description>
            </stat>
            <stat>
                <name>dir_dns_failed</name>
                <value>0</value>
                <flag>a</flag>
                <description>DNS director failed lookups</description>
            </stat>
            <stat>
                <name>dir_dns_hit</name>
                <value>0</value>
                <flag>a</flag>
                <description>DNS director cached lookups hit</description>
            </stat>
            <stat>
                <name>dir_dns_cache_full</name>
                <value>0</value>
                <flag>a</flag>
                <description>DNS director full dnscache</description>
            </stat>
            <stat>
                <name>vmods</name>
                <value>0</value>
                <flag>i</flag>
                <description>Loaded VMODs</description>
            </stat>
            <stat>
                <name>n_gzip</name>
                <value>0</value>
                <flag>a</flag>
                <description>Gzip operations</description>
            </stat>
            <stat>
                <name>n_gunzip</name>
                <value>0</value>
                <flag>a</flag>
                <description>Gunzip operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sms</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sms</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sms</ident>
                <name>locks</name>
                <value>265221</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sms</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smp</ident>
                <name>creat</name>
                <value>0</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smp</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smp</ident>
                <name>locks</name>
                <value>0</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smp</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sma</ident>
                <name>creat</name>
                <value>2</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sma</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sma</ident>
                <name>locks</name>
                <value>74266910</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sma</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smf</ident>
                <name>creat</name>
                <value>0</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smf</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smf</ident>
                <name>locks</name>
                <value>0</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>smf</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hsl</ident>
                <name>creat</name>
                <value>0</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hsl</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hsl</ident>
                <name>locks</name>
                <value>0</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hsl</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcb</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcb</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcb</ident>
                <name>locks</name>
                <value>5991389</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcb</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcl</ident>
                <name>creat</name>
                <value>0</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcl</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcl</ident>
                <name>locks</name>
                <value>0</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>hcl</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vcl</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vcl</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vcl</ident>
                <name>locks</name>
                <value>82036</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vcl</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>stat</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>stat</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>stat</ident>
                <name>locks</name>
                <value>334</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>stat</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sessmem</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sessmem</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sessmem</ident>
                <name>locks</name>
                <value>480320</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>sessmem</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wstat</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wstat</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wstat</ident>
                <name>locks</name>
                <value>1636233</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wstat</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>herder</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>herder</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>herder</ident>
                <name>locks</name>
                <value>35720</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>herder</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wq</ident>
                <name>creat</name>
                <value>2</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wq</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wq</ident>
                <name>locks</name>
                <value>32549458</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>wq</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>objhdr</ident>
                <name>creat</name>
                <value>3056380</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>objhdr</ident>
                <name>destroy</name>
                <value>2936840</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>objhdr</ident>
                <name>locks</name>
                <value>195679071</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>objhdr</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>exp</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>exp</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>exp</ident>
                <name>locks</name>
                <value>22976360</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>exp</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>lru</ident>
                <name>creat</name>
                <value>2</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>lru</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>lru</ident>
                <name>locks</name>
                <value>17201617</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>lru</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>cli</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>cli</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>cli</ident>
                <name>locks</name>
                <value>88310</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>cli</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>ban</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>ban</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>ban</ident>
                <name>locks</name>
                <value>22976610</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>ban</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbp</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbp</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbp</ident>
                <name>locks</name>
                <value>1412365</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbp</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbe</ident>
                <name>creat</name>
                <value>1</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbe</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbe</ident>
                <name>locks</name>
                <value>377647</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>vbe</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>backend</ident>
                <name>creat</name>
                <value>17</value>
                <flag>a</flag>
                <description>Created locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>backend</ident>
                <name>destroy</name>
                <value>0</value>
                <flag>a</flag>
                <description>Destroyed locks</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>backend</ident>
                <name>locks</name>
                <value>52925585</value>
                <flag>a</flag>
                <description>Lock Operations</description>
            </stat>
            <stat>
                <type>LCK</type>
                <ident>backend</ident>
                <name>colls</name>
                <value>0</value>
                <flag>a</flag>
                <description>Collisions</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>c_req</name>
                <value>26057659</value>
                <flag>a</flag>
                <description>Allocator requests</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>c_fail</name>
                <value>732800073660</value>
                <flag>a</flag>
                <description>Allocator failures</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>c_bytes</name>
                <value>506580500540</value>
                <flag>a</flag>
                <description>Bytes allocated</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>c_freed</name>
                <value>500138170202</value>
                <flag>a</flag>
                <description>Bytes freed</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>g_alloc</name>
                <value>791369</value>
                <flag>i</flag>
                <description>Allocations outstanding</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>g_bytes</name>
                <value>6442330338</value>
                <flag>i</flag>
                <description>Bytes outstanding</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>s0</ident>
                <name>g_space</name>
                <value>120606</value>
                <flag>i</flag>
                <description>Bytes available</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>c_req</name>
                <value>12490635</value>
                <flag>a</flag>
                <description>Allocator requests</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>c_fail</name>
                <value>0</value>
                <flag>a</flag>
                <description>Allocator failures</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>c_bytes</name>
                <value>44213239981</value>
                <flag>a</flag>
                <description>Bytes allocated</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>c_freed</name>
                <value>44213169757</value>
                <flag>a</flag>
                <description>Bytes freed</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>g_alloc</name>
                <value>192</value>
                <flag>i</flag>
                <description>Allocations outstanding</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>g_bytes</name>
                <value>70224</value>
                <flag>i</flag>
                <description>Bytes outstanding</description>
            </stat>
            <stat>
                <type>SMA</type>
                <ident>Transient</ident>
                <name>g_space</name>
                <value>0</value>
                <flag>i</flag>
                <description>Bytes available</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>default(127.0.0.1,,80)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>default(127.0.0.1,,80)</ident>
                <name>happy</name>
                <value>0</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_01(10.93.67.16,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_01(10.93.67.16,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_02(10.93.67.17,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_02(10.93.67.17,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_03(10.93.67.18,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_03(10.93.67.18,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_04(10.93.67.105,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_04(10.93.67.105,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_06(10.93.67.133,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_06(10.93.67.133,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_07(10.93.67.134,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_video_prd_services_07(10.93.67.134,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_01(10.93.67.75,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_01(10.93.67.75,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_02(10.93.67.76,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_02(10.93.67.76,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_03(10.93.67.77,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_03(10.93.67.77,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_04(10.93.67.121,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_04(10.93.67.121,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_05(10.93.67.122,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_article_prd_services_05(10.93.67.122,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_01(10.93.66.132,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_01(10.93.66.132,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_02(10.93.66.150,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_02(10.93.66.150,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_03(10.93.67.107,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_03(10.93.67.107,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_04(10.93.67.125,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_04(10.93.67.125,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_05(10.93.67.126,,8080)</ident>
                <name>vcls</name>
                <value>1</value>
                <flag>i</flag>
                <description>VCL references</description>
            </stat>
            <stat>
                <type>VBE</type>
                <ident>media_object_prd_services_05(10.93.67.126,,8080)</ident>
                <name>happy</name>
                <value>18446744073709551615</value>
                <flag>b</flag>
                <description>Happy health probes</description>
            </stat>
        </varnishstat>
        """

        self.config = """
init_config:

instances:
    -   varnishstat: /usr/bin/varnishstat
"""

    def testParsing(self):
        v, instances = get_check('varnish', self.config)
        v._parse_varnishstat(self.v_dump, False)
        metrics = v.get_metrics()
        self.assertEqual([m[2] for m in metrics
                          if m[0] == "varnish.n_waitinglist"][0], 980)
        assert "varnish.fetch_length" not in [m[0] for m in metrics]

        # XML parsing
        v._parse_varnishstat(self.xml_dump, True)
        metrics = v.get_metrics()
        self.assertEqual([m[2] for m in metrics
                          if m[0] == "varnish.SMA.s0.g_space"][0], 120606)
        assert "varnish.SMA.transient.c_bytes" not in [m[0] for m in metrics]

    def testCheck(self):
        v, instances = get_check('varnish', self.config)
        import pprint
        try:
            for i in range(3):
                v.check({"varnishstat": os.popen("which varnishstat").read()[:-1]})
                pprint.pprint(v.get_metrics())
                time.sleep(1)
        except Exception:
            pass

    def testOldConfig(self):
        v, instances = get_check('varnish', self.config)
        config = {
            'varnishstat': '/usr/bin/varnishstat'
        }

        instances = v.parse_agent_config(config)['instances']
        assert instances[0]['varnishstat'] == config['varnishstat']

if __name__ == '__main__':
    unittest.main()
